# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini@gmail.com
# ---------------------------------------------------------------------------------------------------------------

# ---------------------------------------------------------------------------------------------------------------
# Management of exceptions for piwiPre
# ---------------------------------------------------------------------------------------------------------------

# raise PwpError(): prints a message and generate an exception


import inspect
import time
import datetime
import sys
import termcolor
import platform
import os


class PwpLog:
    def __init__(self):
        logfile_name = time.strftime("piwiPre_%Y_%m_%d.log")
        self.cwd = os.getcwd()  # this is read at thr init of the module, so BEFORE any chdir
        self.quiet = "--quiet" in sys.argv
        self.logfile = None if self.quiet else open(logfile_name, "a", encoding="utf-8")
        self.start_time = datetime.datetime.now()
        self.stop_on_warning = False
        self.print_debug = False
        self.colors = True
        self.data = {           # normal logs are not stored
            'info': [],         # logs that are stored for tests
            'Warning': [],      # a problem that do not require stopping the program
            'ERROR': [],        # under normal circumstances, stops the program. Can be trapped for test
        }
        self.started = False
        self.files_processed = {}
        if platform.system() == "Windows":
            os.system('color')

    def start(self):
        if self.started:
            return
        self.started = True
        self.msg(f"---- piwiPre start {self.start_time}")
        self.msg('')
        self.msg(f"Current Working Directory = '{self.cwd}'")
        self.msg(f"System                    = '{platform.system()}'")
        self.msg(f"HOME                      = '{os.path.expanduser('~')}'")
        if platform.system() == "Windows":
            self.msg(f"Exe Install Directory     = '{os.environ['PROGRAMFILES(X86)']}' /piwiPre /ffmpeg")  # noqa
            self.msg(f"Processor Architecture    = '{os.environ['PROCESSOR_ARCHITECTURE']}'")

        self.msg('')

        self.msg('--------------- Help on logs:')
        self.msg('')

        self.msg('LR[LR]')
        self.msg('     L  : Local file')
        self.msg('     R  : Remote file')
        self.msg('')

        self.msg('A[author]')
        self.msg('')

        self.msg('D[date]')
        self.msg('')

        self.msg('Rot[rf]')
        self.msg('   --  : nothing'),
        self.msg('   |-  : FLIP_LEFT_RIGHT'),  # 2
        self.msg('   ^-  : ROTATE_180'),  # 3
        self.msg('   V-  : FLIP_TOP_BOTTOM'),  # 4
        self.msg('   <|  : ROTATE_270'),  # 5
        self.msg('   <-  : ROTATE_270 + flip'),  # 6
        self.msg('   >|  : ROTATE_90 + flip'),  # 7
        self.msg('   >-  : ROTATE_90')
        self.msg('')

        self.msg('meta[diac]: information changed in the metadata of the file (EXIF or IPTC)')  # noqa
        self.msg('     d  : date')
        self.msg('     i  : instructions')
        self.msg('     a  : author')
        self.msg('     c  : copyright')
        self.msg('')

        self.msg('rep[R]  : Representative picture for video')
        self.msg('')

        self.msg('th[STMCI] : thumbnail')       # noqa
        self.msg('     S  : _sq.jpg')
        self.msg('     T  : _th.jpg')
        self.msg('     M  : _me.jpg')
        self.msg('     C  : _cu_250.jpg')
        self.msg('     I  : _index.jpg')
        self.msg('')

        self.msg('rth[stmci] : Remote thumbnail')       # noqa
        self.msg('     s  : _sq.jpg')
        self.msg('     t  : _th.jpg')
        self.msg('     m  : _me.jpg')
        self.msg('     c  : _cu_250.jpg')
        self.msg('     i  : _index.jpg')
        self.msg('')

        self.msg('db[CSWH5GA]: information changed in the piwigo SQL database')  # noqa
        self.msg('     C  : Created in the db')
        self.msg('     S  : Size')
        self.msg('     W  : Width')
        self.msg('     H  : Height')
        self.msg('     5  : md5')
        self.msg('     G  : GPS info')
        self.msg('     A  : Author')
        self.msg('')

        self.msg('Actions:[album filename] ')
        self.msg('  Keep  : file was already in album, keep that value')
        self.msg('  Renam : Rename to a new filename because of conflict, and copy to album')              # noqa
        self.msg('  Copy  : copy file to album, there was no previous file')
        self.msg('  Updat : update the file in album with new value')                                      # noqa
        self.msg('  Clobb : Clobber previous version in album, rename was not allowed')                    # noqa
        self.msg('  DELET : File was deleted from local album because no corresponding remote file')       # noqa
        self.msg('  ABORT : Processing of file is aborted, because no remote file')
        self.msg('')

        self.msg('Back:[backup filename] ')
        self.msg('')
        # self.test_msg("This is a message generated by test harness")
        # self.warning("This is a warning")

    def __del__(self):
        if self.logfile:
            self.logfile.close()
        print('---- piwiPre End ')

    def configure(self, config):
        self.print_debug = config['debug']
        self.stop_on_warning = config['stop-on-warning']
        self.colors = config['enable-colors']

    def reset_data(self):
        self.start_time = datetime.datetime.now()
        self.data = {
            'info': [],
            'Warning': [],
            'ERROR': [],
        }
        old = self.files_processed
        self.files_processed = {}
        return old

    def msg_nb(self, level):
        return len(self.data[level])

    def end(self):
        if self.quiet:
            return
        end = datetime.datetime.now()
        self.msg(f"--- start         = {self.start_time} ---")
        self.msg(f"--- end           = {end} ---")
        self.msg(f"--- duration      = {end - self.start_time}")
        files = 0
        for k in self.files_processed:
            self.msg(f"--- {k:14} = {self.files_processed[k]} ")
            files += self.files_processed[k]
        if files:
            self.msg(f"--- duration/file = {(end - self.start_time) / files}")
        self.msg("------------------------------------")

    def incr_picture(self, category):
        self.files_processed[category] = 1 if category not in self.files_processed \
            else self.files_processed[category] + 1

    def do_msg(self, msg, context=None, level='msg', flush=False, color=None):
        if context is not None:
            if level not in ['debug', 'info'] or self.print_debug:
                print(f"{level:7} {context}", flush=True)
                if self.logfile:
                    print(f"{level:7} {context}", file=self.logfile)

        if level not in ['debug', 'info'] or self.print_debug:
            line = f"{level:7} {msg}"
            if color and self.colors:
                print(termcolor.colored(line, color=color, force_color=True))
            else:
                print(line, flush=True)
            if self.logfile:
                print(line, file=self.logfile)
            if flush:
                sys.stdout.flush()

        if level != 'msg' and level != 'debug':
            self.data[level].append(msg)

        if level == 'Warning' and self.stop_on_warning:
            raise PwpError("Stop on warning")

        if level == 'Warning' and self.msg_nb('Warning') > 20:
            raise PwpError("Too much warnings")

        if level == 'ERROR' and self.msg_nb('ERROR') > 20:
            raise PwpFatal("Too much errors, aborting")

    def warning(self, msg, context=None):        # warning is always kept for test, and ALWAYS printed
        self.do_msg(msg, context=context, level='Warning', flush=True, color="red")

    def msg(self, msg, context=None, color=None):           # msg is NOT kept for test, and ALWAYS printed
        if not self.quiet:
            self.do_msg(msg, context=context, level='msg', color=color)

    def test_msg(self, msg, context=None):
        self.do_msg("    Test: " + msg, context=context, level='msg', color='blue')

    def info(self, msg, context=None):          # info is always kept for test, and printed only if --debug
        self.do_msg(msg, context=context, level='info', flush=True)

    def debug(self, msg, context=None):         # debug is NOT kept for test, and printed only if --debug
        self.do_msg(msg, context=context, level='debug', flush=True)

    # error must NOT be declared, it is mandatory to go through PwpError
    # def error(self, msg, context=None):
    #     self.do_msg(msg, context=context, level='ERROR', flush=True)


class PwpError(Exception):
    def __init__(self, msg, context=None):
        LOGGER.do_msg(msg, level='ERROR', context=context, flush=True, color="magenta")


class PwpInternalError(PwpError):
    def __init__(self, msg: str):  # pragma: no cover
        previous_frame = inspect.currentframe().f_back
        (filename, line_number, _function_name, _lines, _index) = inspect.getframeinfo(previous_frame)
        context = f"{filename}:{line_number:3}"
        super().__init__("INTERNAL " + msg, context=context)


class PwpConfigError(PwpError):
    def __init__(self, msg, context="Cmd-line/configuration "):
        super().__init__(msg, context=context)


# class PwpExit(PwpError):
#     def __init__(self, msg):
#         super().__init__(msg, context="Exiting from main")


class PwpFatal(Exception):  # pragma: no cover      # should not be trapped
    def __init__(self, msg):
        LOGGER.do_msg(msg, level='FATAL', flush=True)


LOGGER = PwpLog()



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------

class PwpLicence:
    def __init__(self):
        pass

    @staticmethod
    def print():
        print("This program and library is licenced under the European Union Public Licence v1.2")
        print("")
        print("The full text of EUPL V1.2 can be found, in several languages at:")              # noqa
        print("https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12")
        print("")
        print("and was published on the Official Journal of the European Union 19.5.2017")
        print("")
        print("                      EUROPEAN UNION PUBLIC LICENCE v. 1.2")
        print("                      EUPL © the European Union 2007, 2016")                      # noqa
        print("")
        print("This European Union Public Licence (the ‘EUPL’) applies to the Work (as defined") # noqa
        print("below) which is provided under the terms of this Licence. Any use of the Work,")
        print("other than as authorised under this Licence is prohibited (to the extent such")
        print("use is covered by a right of the copyright holder of the Work).")
        print("")
        print("The Work is provided under the terms of this Licence when the Licensor (as")
        print("defined below) has placed the following notice immediately following the")
        print("copyright notice for the Work:")
        print("")
        print("        Licensed under the EUPL")                                                             # noqa
        print("")
        print("or has expressed by any other means his willingness to license under the EUPL.")             # noqa
        print("")
        print("1. Definitions")
        print("")
        print("In this Licence, the following terms have the following meaning:")
        print("")
        print("- ‘The Licence’: this Licence.")
        print("")
        print("- ‘The Original Work’: the work or software distributed or communicated by the")
        print("  Licensor under this Licence, available as Source Code and also as Executable")
        print("  Code as the case may be.")
        print("")
        print("- ‘Derivative Works’: the works or software that could be created by the")
        print("  Licensee, based upon the Original Work or modifications thereof. This Licence")
        print("  does not define the extent of modification or dependence on the Original Work")
        print("  required in order to classify a work as a Derivative Work; this extent is")
        print("  determined by copyright law applicable in the country mentioned in Article 15.")
        print("")
        print("- ‘The Work’: the Original Work or its Derivative Works.")
        print("")
        print("- ‘The Source Code’: the human-readable form of the Work which is the most")
        print("  convenient for people to study and modify.")
        print("")
        print("- ‘The Executable Code’: any code which has generally been compiled and which is")
        print("  meant to be interpreted by a computer as a program.")
        print("")
        print("- ‘The Licensor’: the natural or legal person that distributes or communicates")
        print("  the Work under the Licence.")
        print("")
        print("- ‘Contributor(s)’: any natural or legal person who modifies the Work under the")
        print("  Licence, or otherwise contributes to the creation of a Derivative Work.")
        print("")
        print("- ‘The Licensee’ or ‘You’: any natural or legal person who makes any usage of")
        print("  the Work under the terms of the Licence.")
        print("")
        print("- ‘Distribution’ or ‘Communication’: any act of selling, giving, lending,")
        print("  renting, distributing, communicating, transmitting, or otherwise making")
        print("  available, online or offline, copies of the Work or providing access to its")
        print("  essential functionalities at the disposal of any other natural or legal")
        print("  person.")
        print("")
        print("2. Scope of the rights granted by the Licence")
        print("")
        print("The Licensor hereby grants You a worldwide, royalty-free, non-exclusive,")
        print("sub-licensable licence to do the following, for the duration of copyright vested")
        print("in the Original Work:")
        print("")
        print("- use the Work in any circumstance and for all usage,")
        print("- reproduce the Work,")
        print("- modify the Work, and make Derivative Works based upon the Work,")
        print("- communicate to the public, including the right to make available or display")
        print("  the Work or copies thereof to the public and perform publicly, as the case may")
        print("  be, the Work,")
        print("- distribute the Work or copies thereof,")
        print("- lend and rent the Work or copies thereof,")
        print("- sublicense rights in the Work or copies thereof.")
        print("")
        print("Those rights can be exercised on any media, supports and formats, whether now")
        print("known or later invented, as far as the applicable law permits so.")
        print("")
        print("In the countries where moral rights apply, the Licensor waives his right to")
        print("exercise his moral right to the extent allowed by law in order to make effective")
        print("the licence of the economic rights here above listed.")
        print("")
        print("The Licensor grants to the Licensee royalty-free, non-exclusive usage rights to")
        print("any patents held by the Licensor, to the extent necessary to make use of the")
        print("rights granted on the Work under this Licence.")
        print("")
        print("3. Communication of the Source Code")
        print("")
        print("The Licensor may provide the Work either in its Source Code form, or as")
        print("Executable Code. If the Work is provided as Executable Code, the Licensor")
        print("provides in addition a machine-readable copy of the Source Code of the Work")
        print("along with each copy of the Work that the Licensor distributes or indicates, in")
        print("a notice following the copyright notice attached to the Work, a repository where")
        print("the Source Code is easily and freely accessible for as long as the Licensor")
        print("continues to distribute or communicate the Work.")
        print("")
        print("4. Limitations on copyright")
        print("")
        print("Nothing in this Licence is intended to deprive the Licensee of the benefits from")
        print("any exception or limitation to the exclusive rights of the rights owners in the")
        print("Work, of the exhaustion of those rights or of other applicable limitations")
        print("thereto.")
        print("")
        print("5. Obligations of the Licensee")
        print("")
        print("The grant of the rights mentioned above is subject to some restrictions and")
        print("obligations imposed on the Licensee. Those obligations are the following:")
        print("")
        print("Attribution right: The Licensee shall keep intact all copyright, patent or")
        print("trademarks notices and all notices that refer to the Licence and to the")
        print("disclaimer of warranties. The Licensee must include a copy of such notices and a")
        print("copy of the Licence with every copy of the Work he/she distributes or")
        print("communicates. The Licensee must cause any Derivative Work to carry prominent")
        print("notices stating that the Work has been modified and the date of modification.")
        print("")
        print("Copyleft clause: If the Licensee distributes or communicates copies of the")
        print("Original Works or Derivative Works, this Distribution or Communication will be")
        print("done under the terms of this Licence or of a later version of this Licence")
        print("unless the Original Work is expressly distributed only under this version of the")
        print("Licence — for example by communicating ‘EUPL v. 1.2 only’. The Licensee")                 # noqa
        print("(becoming Licensor) cannot offer or impose any additional terms or conditions on")
        print("the Work or Derivative Work that alter or restrict the terms of the Licence.")
        print("")
        print("Compatibility clause: If the Licensee Distributes or Communicates Derivative")
        print("Works or copies thereof based upon both the Work and another work licensed under")
        print("a Compatible Licence, this Distribution or Communication can be done under the")
        print("terms of this Compatible Licence. For the sake of this clause, ‘Compatible")
        print("Licence’ refers to the licences listed in the appendix attached to this Licence.")
        print("Should the Licensee's obligations under the Compatible Licence conflict with")
        print("his/her obligations under this Licence, the obligations of the Compatible")
        print("Licence shall prevail.")
        print("")
        print("Provision of Source Code: When distributing or communicating copies of the Work,")
        print("the Licensee will provide a machine-readable copy of the Source Code or indicate")
        print("a repository where this Source will be easily and freely available for as long")
        print("as the Licensee continues to distribute or communicate the Work.")
        print("")
        print("Legal Protection: This Licence does not grant permission to use the trade names,")
        print("trademarks, service marks, or names of the Licensor, except as required for")
        print("reasonable and customary use in describing the origin of the Work and")
        print("reproducing the content of the copyright notice.")
        print("")
        print("6. Chain of Authorship")
        print("")
        print("The original Licensor warrants that the copyright in the Original Work granted")
        print("hereunder is owned by him/her or licensed to him/her and that he/she has the")
        print("power and authority to grant the Licence.")
        print("")
        print("Each Contributor warrants that the copyright in the modifications he/she brings")
        print("to the Work are owned by him/her or licensed to him/her and that he/she has the")
        print("power and authority to grant the Licence.")
        print("")
        print("Each time You accept the Licence, the original Licensor and subsequent")
        print("Contributors grant You a licence to their contributions to the Work, under the")
        print("terms of this Licence.")
        print("")
        print("7. Disclaimer of Warranty")
        print("")
        print("The Work is a work in progress, which is continuously improved by numerous")
        print("Contributors. It is not a finished work and may therefore contain defects or")
        print("‘bugs’ inherent to this type of development.")
        print("")
        print("For the above reason, the Work is provided under the Licence on an ‘as is’ basis")
        print("and without warranties of any kind concerning the Work, including without")
        print("limitation merchantability, fitness for a particular purpose, absence of defects")
        print("or errors, accuracy, non-infringement of intellectual property rights other than")
        print("copyright as stated in Article 6 of this Licence.")
        print("")
        print("This disclaimer of warranty is an essential part of the Licence and a condition")
        print("for the grant of any rights to the Work.")
        print("")
        print("8. Disclaimer of Liability")
        print("")
        print("Except in the cases of wilful misconduct or damages directly caused to natural")
        print("persons, the Licensor will in no event be liable for any direct or indirect,")
        print("material or moral, damages of any kind, arising out of the Licence or of the use")
        print("of the Work, including without limitation, damages for loss of goodwill, work")
        print("stoppage, computer failure or malfunction, loss of data or any commercial")
        print("damage, even if the Licensor has been advised of the possibility of such damage.")
        print("However, the Licensor will be liable under statutory product liability laws as")
        print("far such laws apply to the Work.")
        print("")
        print("9. Additional agreements")
        print("")
        print("While distributing the Work, You may choose to conclude an additional agreement,")
        print("defining obligations or services consistent with this Licence. However, if")
        print("accepting obligations, You may act only on your own behalf and on your sole")
        print("responsibility, not on behalf of the original Licensor or any other Contributor,")
        print("and only if You agree to indemnify, defend, and hold each Contributor harmless")
        print("for any liability incurred by, or claims asserted against such Contributor by")
        print("the fact You have accepted any warranty or additional liability.")
        print("")
        print("10. Acceptance of the Licence")
        print("")
        print("The provisions of this Licence can be accepted by clicking on an icon ‘I agree’")
        print("placed under the bottom of a window displaying the text of this Licence or by")
        print("affirming consent in any other similar way, in accordance with the rules of")
        print("applicable law. Clicking on that icon indicates your clear and irrevocable")
        print("acceptance of this Licence and all of its terms and conditions.")
        print("")
        print("Similarly, you irrevocably accept this Licence and all of its terms and")
        print("conditions by exercising any rights granted to You by Article 2 of this Licence,")
        print("such as the use of the Work, the creation by You of a Derivative Work or the")
        print("Distribution or Communication by You of the Work or copies thereof.")
        print("")
        print("11. Information to the public")
        print("")
        print("In case of any Distribution or Communication of the Work by means of electronic")
        print("communication by You (for example, by offering to download the Work from a")
        print("remote location) the distribution channel or media (for example, a website) must")
        print("at least provide to the public the information requested by the applicable law")
        print("regarding the Licensor, the Licence and the way it may be accessible, concluded,")
        print("stored and reproduced by the Licensee.")
        print("")
        print("12. Termination of the Licence")
        print("")
        print("The Licence and the rights granted hereunder will terminate automatically upon")
        print("any breach by the Licensee of the terms of the Licence.")
        print("")
        print("Such a termination will not terminate the licences of any person who has")
        print("received the Work from the Licensee under the Licence, provided such persons")
        print("remain in full compliance with the Licence.")
        print("")
        print("13. Miscellaneous")
        print("")
        print("Without prejudice of Article 9 above, the Licence represents the complete")
        print("agreement between the Parties as to the Work.")
        print("")
        print("If any provision of the Licence is invalid or unenforceable under applicable")
        print("law, this will not affect the validity or enforceability of the Licence as a")
        print("whole. Such provision will be construed or reformed so as necessary to make it")
        print("valid and enforceable.")
        print("")
        print("The European Commission may publish other linguistic versions or new versions of")
        print("this Licence or updated versions of the Appendix, so far this is required and")
        print("reasonable, without reducing the scope of the rights granted by the Licence. New")
        print("versions of the Licence will be published with a unique version number.")
        print("")
        print("All linguistic versions of this Licence, approved by the European Commission,")
        print("have identical value. Parties can take advantage of the linguistic version of")
        print("their choice.")
        print("")
        print("14. Jurisdiction")
        print("")
        print("Without prejudice to specific agreement between parties,")
        print("")
        print("- any litigation resulting from the interpretation of this License, arising")
        print("  between the European Union institutions, bodies, offices or agencies, as a")
        print("  Licensor, and any Licensee, will be subject to the jurisdiction of the Court")
        print("  of Justice of the European Union, as laid down in article 272 of the Treaty on")
        print("  the Functioning of the European Union,")
        print("")
        print("- any litigation arising between other parties and resulting from the")
        print("  interpretation of this License, will be subject to the exclusive jurisdiction")
        print("  of the competent court where the Licensor resides or conducts its primary")
        print("  business.")
        print("")
        print("15. Applicable Law")
        print("")
        print("Without prejudice to specific agreement between parties,")
        print("")
        print("- this Licence shall be governed by the law of the European Union Member State")
        print("  where the Licensor has his seat, resides or has his registered office,")
        print("")
        print("- this licence shall be governed by Belgian law if the Licensor has no seat,")
        print("  residence or registered office inside a European Union Member State.")
        print("")
        print("Appendix")
        print("")
        print("‘Compatible Licences’ according to Article 5 EUPL are:")                      # noqa
        print("")
        print("- GNU General Public License (GPL) v. 2, v. 3")
        print("- GNU Affero General Public License (AGPL) v. 3")                             # noqa
        print("- Open Software License (OSL) v. 2.1, v. 3.0")
        print("- Eclipse Public License (EPL) v. 1.0")
        print("- CeCILL v. 2.0, v. 2.1")                                                         # noqa
        print("- Mozilla Public Licence (MPL) v. 2")
        print("- GNU Lesser General Public Licence (LGPL) v. 2.1, v. 3")
        print("- Creative Commons Attribution-ShareAlike v. 3.0 Unported (CC BY-SA 3.0) for")               # noqa
        print("  works other than software")
        print("- European Union Public Licence (EUPL) v. 1.1, v. 1.2")                               # noqa
        print("- Québec Free and Open-Source Licence — Reciprocity (LiLiQ-R) or Strong")                         # noqa
        print("  Reciprocity (LiLiQ-R+).")
        print("")
        print("The European Commission may update this Appendix to later versions of the above")
        print("licences without producing a new version of the EUPL, as long as they provide")                   # noqa
        print("the rights granted in Article 2 of this Licence and protect the covered Source")
        print("Code from exclusive appropriation.")
        print("")
        print("All other changes or additions to this Appendix require the production of a new")
        print("EUPL version.")                                                                             # noqa
    # End of patched text



import os
import datetime

from piwiPre.pwpActor import ACTOR
from piwiPre.pwpConfig import PwpConfig
from piwiPre.pwpErrors import PwpInternalError, LOGGER

# DONE:
#  Optimize md5sum by caching pwpDirEntry in a class variable
#  cache is reset at the end of run_stage_dir
#  files created/removed must be reflected in cache
#  A method to achieve this is to perform ALL actions through PwpFileEntry


class ItemEntry:
    def __init__(self, local: str or None, is_file: bool):
        self.basename = os.path.basename(local) if local else None
        self.is_file = is_file      # file = True, else  dir

        self.local = local          # local is the theoretical local name
        self.is_local = False       # is_local means that the file/dir DOES exist

        self.remote = None          # remote is the theoretical remote name
        self.is_remote = False      # is_remote means that the file/dir DOES exist

    def remove(self, father: 'ItemEntry' or None, msg: str = ''):
        raise PwpInternalError(f"ItemEntry.remove({self})")

    def set_remote(self, remote):
        self.remote = remote
        self.basename = os.path.basename(remote)
        self.is_remote = True

    def set_local(self, local):
        self.local = local
        self.basename = os.path.basename(local)
        self.is_local = True


class PwpDirEntry(ItemEntry):
    cache = {}

    def __init__(self, local: str, config: PwpConfig, context: str):
        """
        PwpDirEntry(self, local: str, config: PwpConfig, context: str)
        :param local: dir path in the local filesystem
        :param config: config
        :param context: either local, web or album.
        """
        super().__init__(local=local, is_file=False)
        self.sons: dict[str, PwpDirEntry] = {}
        self.files: dict[str, PwpFileEntry] = {}
        self.config = config
        self.context = context

        self.is_local = os.path.isdir(local)
        self.is_opened = False

        if context == 'album' and ACTOR.ssh_connection and config['remote-album'] and config['enable-remote-album']:
            if not local.startswith(config['album']):
                raise PwpInternalError(f"Illegal dir path '{local}' does not start with album '{config['album']}'")
            self.remote = local.replace(config['album'], config['remote-album'], 1)

        elif context == 'web' and ACTOR.ssh_connection and config['remote-web'] and config['enable-remote-web']:
            if not local.startswith(config['web']):
                raise PwpInternalError(f"Illegal dir path {local} does not start with web '{config['web']}'")
            self.remote = local.replace(config['web'], config['remote-web'], 1)

        if self.basename is None:
            if self.remote is None:
                raise PwpInternalError("PwpDirEntry(None, None")
            self.basename = os.path.basename(self.remote)

        if self.local in PwpDirEntry.cache:
            raise PwpInternalError(f"Duplicate DirEntry {self.local}")
        PwpDirEntry.cache[self.local] = self

    @staticmethod
    def clean_cache():
        PwpDirEntry.cache = {}
        # LOGGER.msg("clean_cache")

    def all_entries(self):
        return list(self.sons.values()) + list(self.files.values())

    def is_empty(self):
        return len(self.files) + len(self.sons) == 0

    def remove(self, father: 'ItemEntry' or None, msg: str = ''):
        if self.local:
            ACTOR.rmtree(self.local, msg=msg)
            if msg:
                LOGGER.msg(f"{msg} : local dir {self.local}")
        if self.remote:
            ACTOR.remote_rmtree(self.remote, msg=msg)
            if msg:
                LOGGER.msg(f"{msg} : remote dir {self.remote}")
        if father:
            father.sons.pop(self.basename)

    # @staticmethod
    # def create(local: str, config: PwpConfig, context: str):
    #     if local in PwpDirEntry.cache:
    #         return PwpDirEntry.cache[local]
    #     return PwpDirEntry(local, config, context)

    @staticmethod
    def open(local: str, config: PwpConfig, context: str):
        if local in PwpDirEntry.cache:
            res = PwpDirEntry.cache[local]
        else:
            res = PwpDirEntry(local, config, context)
        res.read()
        return res

    def read(self):
        if not self.is_opened:
            self.sons: dict[str, PwpFileEntry] = {}
            self.files: dict[str, PwpFileEntry] = {}
            if os.path.isdir(self.local):
                all_files = os.listdir(self.local)
                basename: str
                for basename in all_files:
                    local_path = self.local + '/' + basename
                    # There is no reason why item would be already created
                    if ACTOR.isfile(local_path):
                        existing = self.files[basename] if basename in self.files else None
                        this_file = existing or PwpFileEntry(local_path,
                                                             config=self.config, context=self.context,
                                                             father=self)
                        this_file.set_local(local_path)
                        self.files[basename] = this_file
                        # FileEntry always check for local presence of the file
                    else:
                        existing = self.sons[basename] if basename in self.sons else None
                        this_dir = existing or PwpDirEntry(local_path, self.config, self.context)
                        this_dir.set_local(local_path)
                        self.sons[basename] = this_dir
                        # DirDescr always check for local presence of the file
            self.is_opened = True

            if self.remote:
                all_r_files = ACTOR.remote_ls(self.remote, forced=True, warn_if_absent=False)

                for basename, dico in all_r_files.items():
                    local_path = self.local + '/' + basename if self.local else None
                    remote_path = self.remote + '/' + basename

                    if dico['type'] == 'file':
                        existing = self.files[basename] if basename in self.files else None
                        this_file = existing or PwpFileEntry(local_path,
                                                             config=self.config, context=self.context,
                                                             father=self)
                        this_file.set_remote(remote_path)
                        this_file.remote_size = dico['size']
                        this_file.remote_mdt = dico['date']
                        self.files[basename] = this_file
                    else:  # must be a directory
                        existing = self.sons[basename] if basename in self.sons else None
                        if existing:
                            this_dir = existing
                        else:
                            ACTOR.mkdirs(local_path)
                            this_dir = PwpDirEntry(local_path, self.config, self.context)
                        this_dir.set_local(local_path)
                        this_dir.set_remote(remote_path)
                        self.sons[basename] = this_dir
            LOGGER.info(f"read({self.local, self.remote})")

        def get_base(item):
            return item.basename if item is not None else None

        all_files = sorted(self.files.values(), key=get_base)
        all_sons = sorted(self.sons.values(), key=get_base)
        return all_files, all_sons

    def exists_and_younger_than(self, filename, mdt: datetime.datetime):
        if filename not in self.files:
            return False
        return self.files[filename].is_younger_than(mdt)


class PwpFileEntry(ItemEntry):
    def __init__(self, local: str, context: str, config: PwpConfig, father: PwpDirEntry or None = None):
        """

        :param local: local path, or None
        """

        super().__init__(local=local, is_file=True)
        self.size = None
        self.mdt = None
        self.md5sum = None
        if father is None:
            father = PwpDirEntry.open(os.path.dirname(local), config=config, context=context)
        self.father = father
        self.config = config

        basename = os.path.basename(self.local)
        if basename in father.files:
            raise PwpInternalError(f'duplicate file {self.father}')

        father.files[basename] = self

        if ACTOR.isfile(local):
            self.is_local = True
            self.size = os.stat(local).st_size
            self.mdt = datetime.datetime.fromtimestamp(os.path.getmtime(local))
        else:
            self.is_local = False
            self.size = None
            self.mdt = None

        self.remote = None
        if ACTOR.ssh_connection and self.config['enable-remote-web'] and self.local.startswith(self.config['web']):
            self.remote = self.local.replace(self.config['web'], self.config['remote-web'])
        elif (ACTOR.ssh_connection and
              self.config['enable-remote-album'] and
              self.local.startswith(self.config['album'])):
            self.remote = self.local.replace(self.config['album'], self.config['remote-album'])

        self.remote_size = None
        self.remote_mdt = None
        self.remote_md5sum = None

        if self.basename is None:
            if self.remote is None:
                raise PwpInternalError("PwpFileEntry(None, None")
            self.basename = os.path.basename(self.remote)

    def exists(self):
        if self.is_local and not self.remote:
            # we are in a purely local configuration, otherwise remote would be set
            return True

        if self.remote and self.is_remote:
            # we are in a remote configuration, only is_remote is important, we do not care about is_local
            return True

        return False

    @staticmethod
    def lookup(local: str, context: str, config: PwpConfig):
        father = PwpDirEntry.open(os.path.dirname(local), config=config, context=context)
        if os.path.basename(local) in father.files:
            return father.files[os.path.basename(local)]
        return PwpFileEntry(local, context=context, config=config)

    def is_younger_than(self, mdt: datetime.datetime):
        if self.is_local:
            return self.mdt >= mdt
        # if there is only a remote file, we are paranoiac and create again the thumbnail
        return False

    def remove(self, father: 'ItemEntry' or None, msg: str = ''):
        if self.is_local:
            ACTOR.delete(self.local, msg=msg)
            if msg:
                LOGGER.msg(f"delete file {self.local} {msg} ")
        if self.is_remote:
            ACTOR.remote_delete(self.remote, msg=msg)
            if msg:
                LOGGER.msg(f"remote delete file{self.remote} {msg} ")
        if father:
            father.files.pop(self.basename)

    def get_remote_md5(self):
        if not self.is_remote:
            return None
        if not self.remote_md5sum:
            sql_file_info = ACTOR.sql_get_file_info(self.remote)
            if sql_file_info and sql_file_info.md5sum:
                self.remote_md5sum = sql_file_info.md5sum
                LOGGER.test_msg(f"file '{self.local}' in database without md5sum")
                # piwiPre always insert md5sum in the database, and this is enforced when doing verify-album
            else:
                self.remote_md5sum = ACTOR.remote_compute_md5(self.remote)
        return self.remote_md5sum

    def get_md5(self):
        if self.is_local and self.md5sum is None:
            self.md5sum = ACTOR.compute_md5(self.local)
        if self.is_remote:
            self.get_remote_md5()

    def remote_equals_local(self):
        self.get_md5()
        return self.md5sum == self.remote_md5sum

    #       When the same file exists in album and remote-album,
    #       piwiPre verifies that it is actually the same file by comparing the md5 sums
    #       if different, the local file is clobbered by a copy of the remote file

    def get(self):
        """
        get the file from remote location if needed.
        :return: True if the file is really here, maybe after remote-get
                 False if the file is NOT here,
                 The only case when this should happen is --dryrun
        """
        if not self.is_remote:
            return True
        if self.is_local and self.remote_equals_local():
            return True
        if ACTOR.dryrun:
            LOGGER.msg(f"Would have get {self.local} from remote location")
            return False

        if self.is_local:
            LOGGER.warning(f"Incoherent local file '{self.local}' refresh it from remote location")
        ACTOR.remote_get(self.remote, self.local)

        self.is_local = True
        self.md5sum = self.remote_md5sum
        return True

    def put(self):
        if self.remote:
            ACTOR.remote_put(self.local, os.path.dirname(self.remote))
            self.is_remote = True
            self.remote_md5sum = self.md5sum
            return True
        return False

    def synchronize(self):
        if self.is_local and self.is_remote:
            return
        if self.is_local:
            self.put()
            return
        if self.is_remote:
            self.get()

    def local_coherent_with_remote(self):
        """
        local_coherent_with_remote() : detects incoherent local file and deletes it.
        - if there is only one remote file, there is no conflict.
        - if there is a local file without a remote file, this is a violation of assert 2), a warning is raised and
          the local file is ignored.
        - if md5(local) == md5(remote), then it is the same file, and there is no conflict between them
        - if there are 2 files with a different md5, then piwipre considers that the remote version is valid,
          then a warning is raised and the local version is ignored
        :return: True if local file is normal, False if it is incoherent and removed.
        """

        if self.is_local:
            if self.is_remote:
                if self.remote_equals_local():
                    return True
                self.is_local = False
                LOGGER.warning(f"Incoherent local and remote {self.local}: ignore local version")
                return False

            if self.remote:
                # so enable-remote is True, we SHOULD have a remote file, but self.is_remote is false
                self.is_local = False
                LOGGER.warning(f"Local file '{self.local}' without remote : ignore local version")
                return False

        return True



# ===============================================
# Disclaimer:
# This code is a pure copy of elevate-0.1.3
# as found on https://pypi.org/project/elevate
# and maintained by barney.gale@gmail.com
#
# copied here because elevate refuses to install on my PyCharm setup, with strange windows error
# ===============================================

import ctypes
import os
from ctypes import POINTER, c_ulong, c_char_p, c_int, c_void_p
from ctypes.wintypes import HANDLE, BOOL, DWORD, HWND, HINSTANCE, HKEY
from ctypes import windll
import subprocess
import sys

# Constant definitions

SEE_MASK_NOCLOSEPROCESS = 0x00000040
SEE_MASK_NO_CONSOLE = 0x00008000

# Type definitions

PHANDLE = ctypes.POINTER(HANDLE)
PDWORD = ctypes.POINTER(DWORD)


class ShellExecuteInfo(ctypes.Structure):
    _fields_ = [
        ('cbSize',       DWORD),
        ('fMask',        c_ulong),
        ('hwnd',         HWND),
        ('lpVerb',       c_char_p),
        ('lpFile',       c_char_p),
        ('lpParameters', c_char_p),
        ('lpDirectory',  c_char_p),
        ('nShow',        c_int),
        ('hInstApp',     HINSTANCE),
        ('lpIDList',     c_void_p),
        ('lpClass',      c_char_p),
        ('hKeyClass',    HKEY),
        ('dwHotKey',     DWORD),
        ('hIcon',        HANDLE),
        ('hProcess',     HANDLE)]

    def __init__(self, **kw):
        super(ShellExecuteInfo, self).__init__()
        self.cbSize = ctypes.sizeof(self)
        for field_name, field_value in kw.items():
            setattr(self, field_name, field_value)


PShellExecuteInfo = POINTER(ShellExecuteInfo)


# Function definitions

ShellExecuteEx = windll.shell32.ShellExecuteExA
ShellExecuteEx.argtypes = (PShellExecuteInfo, )
ShellExecuteEx.restype = BOOL

WaitForSingleObject = windll.kernel32.WaitForSingleObject
WaitForSingleObject.argtypes = (HANDLE, DWORD)
WaitForSingleObject.restype = DWORD

CloseHandle = windll.kernel32.CloseHandle
CloseHandle.argtypes = (HANDLE, )
CloseHandle.restype = BOOL


# At last, the actual implementation!

# patch: need to pass CWD
#        in case there was a chdir AFTER the start of the initial program
#        the new elevated instance would NOT start on the same dir.
# patch: execute the SAME python interpreter with the same cmdline arguments
#        so rely on orig_argv rather than argv
# patch: need to restaure the path as it was for the original program

def find_in_path(exe):
    for item in os.environ['PATH']:
        tried = item + '/' + exe
        if os.path.isfile(tried):
            return tried
    return None


def elevate(cwd, show_console=True, graphical=True):
    if windll.shell32.IsUserAnAdmin():
        return
    os.chdir(cwd)
    params = ShellExecuteInfo(
        fMask=SEE_MASK_NOCLOSEPROCESS | SEE_MASK_NO_CONSOLE,
        hwnd=None,
        lpVerb=b'runas',
        # lpDirectory=cwd,  # attention: / et \
        # lpFile=sys.executable.encode('cp1252'),
        lpFile=sys.orig_argv[0].encode('cp1252'),
        # lpParameters=subprocess.list2cmdline(sys.argv).encode('cp1252'),
        lpParameters=subprocess.list2cmdline(sys.orig_argv[1:]).encode('cp1252'),
        nShow=int(show_console))
    # print(f"sys.executable = {sys.executable}")
    # print(f"sys.argv       = {sys.argv}")
    # print(f"sys.orig_argv  = {sys.orig_argv}")
    # input("msg:\n")
    try:
        if not ShellExecuteEx(ctypes.byref(params)):
            data = input("Please enter the message 20:\n")
            raise ctypes.WinError()
    except OSError:
        data = input("Please enter the message 2:\n")

    handle = params.hProcess

    ret = DWORD()
    WaitForSingleObject(handle, -1)
    if windll.kernel32.GetExitCodeProcess(handle, ctypes.byref(ret)) == 0:
        raise ctypes.WinError()
    CloseHandle(handle)
    sys.exit(ret.value)



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------

import os
import datetime
# pip install pyyaml
import yaml
import yaml.parser

from piwiPre.pwpActor import ACTOR
from piwiPre.pwpErrors import PwpInternalError, PwpConfigError, LOGGER

# DONE: specifications are imported form code, not from doc
# DONE: parse also the list of possible values, so that we can do a verification


class PwpConfig(dict):
    format_dico = {
        'author': '{author}',
        'base': '{base}',
        'Y': '{Y:04}',
        'm': '{m:02}',
        'd': '{d:02}',
        'H': '{H:02}',
        'M': '{M:02}',
        'S': '{S:02}',
        'month_name': '{month_name}',
        'count': '{count:03}',
        'suffix': '{suffix}',
    }
    legal_items = None

    def __init__(self, content_str: str = None, filename: str = None, dico=None):
        super().__init__()
        LOGGER.msg(f"Reading configuration from '{filename}'")
        self.format_cached = {}

        if content_str is not None:
            try:
                dico = yaml.safe_load(content_str)
            except yaml.parser.ParserError as error:
                error: yaml.parser.ParserError
                context = f"in file {filename}: line : {error.problem_mark.line}"
                msg = "Yaml parsing error : " + error.args[0] + ' ' + error.args[2]
                raise PwpConfigError(msg, context=context)
        elif type(dico) is not dict:
            raise PwpInternalError("illegal PwpConfig")

        if dico is None:
            dico = {}

        dico['ini-filename-parsed'] = filename
        # postprocessing
        for key, value in dico.items():
            self[PwpConfig.normalize(key)] = PwpConfig.normalize(value)

        if PwpConfig.legal_items is None:
            PwpConfig.legal_items = list(self.keys())   # better to do an explicit copy here

    @staticmethod
    def normalize(val):
        if val == 'True' or val == 'true':
            return True
        if val == 'False' or val == 'false':
            return False
        if val == 'None' or val == 'none':
            return None
        if val is None or val is True or val is False:
            return val
        if val == '':
            return val
        if isinstance(val, str) and ((val[0] == "'" and val[-1] == "'") or (val[0] == '"' and val[-1] == '"')) :
            return PwpConfig.normalize(val[1:-1])
        if isinstance(val, int):
            return val
        if isinstance(val, str):
            if val[-1:] == '/':  # remove trailing / for ALL items.
                return PwpConfig.normalize(val[:-1])
            else:
                return val
        if isinstance(val, list):
            nl = [PwpConfig.normalize(x) for x in val]
            return nl
        if isinstance(val, dict):
            nd = {}
            for k, v in val.items():
                nd[PwpConfig.normalize(k)] = PwpConfig.normalize(v)
            return nd
        raise PwpInternalError("Normalize illegal type")

    @staticmethod
    def parse_ini_file(filename):
        if not ACTOR.isfile(filename):
            return PwpConfig(content_str="", filename=filename)

        with ACTOR.open(filename, "r") as ini:
            content = ini.readlines()
        content_str = "".join(content)
        conf = PwpConfig(content_str=content_str, filename=os.path.abspath(filename))
        legals = PwpConfig.legal_items
        for key, value in conf.items():
            if key not in legals:
                raise PwpConfigError(f"Illegal configuration item '{key} : {value}' in '{filename}'")
            conf[key] = None if value == 'None' else value
        return conf

    def merge_ini(self, old):
        LOGGER.debug(f"merging ini files '{old['ini-filename-parsed']}' and '{self['ini-filename-parsed']}'")
        for key in self.keys():
            if key not in old:
                # the old is the default, so it has all the keys.
                raise PwpConfigError(f"ERROR: illegal argument {key}")

        for key, value in old.items():
            if key not in self:
                self[key] = value

            # otherwise, we keep the new value.
        return self

    @staticmethod
    def args_to_dict(args):
        args_dict = vars(args)

        for key, value in args_dict.items():
            args_dict[key] = PwpConfig.normalize(value)

        return args_dict

    def merge_ini_args(self, args, arguments: list):
        """
        merges self and args
        :param args: arguments after parsing by argparse, takes default value into account
        :param arguments: argument list BEFORE parsing by argparse
        :return: self
        """
        LOGGER.debug("merging ini with cmdline args")

        args_dict = self.args_to_dict(args)

        for key, value in self.items():
            flag = key.replace('-', '_')
            expected = '--' + key
            if flag in args_dict and expected in arguments:
                # if key not in arguments, then args contains the default value
                # so the .ini file has higher priority
                self[key] = args_dict[flag]  # or value

        # manage items that are in args but not in config (aka self)
        for flag in args_dict.keys():
            key = flag.replace('_', '-')
            if key not in self:
                self[key] = args_dict[flag]
        return self

    def push_local_ini(self, filename):
        if ACTOR.isfile(filename):
            new_ini = PwpConfig.parse_ini_file(filename)
            new_ini.merge_ini(self)
            return new_ini
        return self

    def author(self, apn, _date):
        author = 'Photographer'
        authors = self['authors']
        if apn in authors:
            author = authors[apn]
        elif 'DEFAULT' in authors:
            author = authors['DEFAULT']
        LOGGER.info(f"apn '{apn}'  => author '{author}'")
        return author

    @staticmethod
    def absolute_date(photo_date: datetime, absolute: dict):

        absolute['hour'] = photo_date.hour if 'hour' not in absolute and photo_date else 0
        absolute['minute'] = photo_date.minute if 'minute' not in absolute and photo_date else 0
        absolute['second'] = photo_date.second if 'second' not in absolute and photo_date else 0

        LOGGER.debug(f"absolute-date {absolute}")
        abs_datetime = datetime.datetime(**absolute)
        return abs_datetime

    def fix_date(self, filename, photo_date, apn):
        all_dates = self['dates']
        if not isinstance(all_dates, dict):
            all_dates = {}

        if photo_date is None:
            if 'NO-DATE' in all_dates:
                nd = all_dates['NO-DATE']
                if 'forced' in nd:
                    return self.absolute_date(photo_date, nd['forced'])
                raise PwpConfigError(f"'{filename}' NO-DATE statement without a 'forced: date'")

            LOGGER.debug(f"'{filename}' without a date and without a correction: no NO-DATE valid statement")
            return None

        for key, descr in all_dates.items():
            if key == 'NO-DATE':
                continue
            start = datetime.datetime(**descr['start']) if 'start' in descr else None
            end = datetime.datetime(**descr['end']) if 'end' in descr else None

            found_apn = apn if apn in descr else 'default' if 'default' in descr else None
            if found_apn and (key == 'ALL' or (start and end and start <= photo_date <= end)):
                update = descr[found_apn]
                if 'delta' in update:
                    new_date = photo_date + datetime.timedelta(**update['delta'])
                    LOGGER.msg(f"DATE correction: {filename}:{apn} (delta) {photo_date} -> {new_date}")
                    return new_date
                if 'forced' in update:
                    nd = update['forced'].copy()
                    new_date = self.absolute_date(photo_date, nd)
                    LOGGER.msg(f"DATE correction: {filename}:{apn} (forced) {photo_date} -> {new_date}")
                    return new_date

                LOGGER.warning(f"date correction start:{start} end:{end} camera:{apn} " +
                               "without a delta or forced statement")

        return photo_date

    def format(self, field):
        if field not in self.format_cached:
            self.format_cached[field] = self[field].format(**self.format_dico)
        return self.format_cached[field]

    def format_dict(self, date, author, base='', count=1, suffix='.jpg'):
        """
        :param date: inherited from the IPTC date of the picture.
        :param author: picture author from IPTC data
        :param base: is the name of the TRIAGE folder where the picture was originally found.
        :param count:
        :param suffix: file suffix
        :return: the dictionary used to format a file or document
        """
        month = self['month-name']
        month_name = month[date.month-1]
        dico = {
            'author': author,
            'base': base,
            'Y': date.year,
            'm': date.month,
            'd': date.day,
            'H': date.hour,
            'M': date.minute,
            'S': date.second,
            'month_name': month_name,
            'count': count,
            'suffix': suffix,
        }
        return dico



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------

import sys
import os.path
import re
import pprint
import tempfile
import datetime

from piwiPre.pwpActor import ACTOR, FileInfo, DirInfo, PwpSummary
from piwiPre.pwpParser import PwpParser
from piwiPre.pwpConfig import PwpConfig
from piwiPre.pwpData import PwpJpg, PwpData, PwpVideo
from piwiPre.pwpDir import PwpDirEntry, PwpFileEntry
from piwiPre.pwpErrors import PwpError, PwpConfigError, PwpInternalError, LOGGER
from piwiPre.pwpVersion import PwpVersion
from piwiPre.pwpLicence import PwpLicence

# -----------------------------------------------------------------------------------
# Requirements, General behavior
#
# REQ 0001: piwiPre is configured with HOME/.piwiPre.ini and piwiPre.ini files found in the hierarchy of directories
# REQ 0002: piwiPre is also configured by cmdline arguments
# REQ 0003: piwiPre renames .jpg, .mp4, .txt files found in 'triage' (enable-rename) (program_24)
# REQ 0004: piwiPre inserts metadata in .jpg and .mp4 files (enable-metadata)
# REQ 0005: piwiPre generates piwigo metadata (enable-thumbnails)
# REQ 0006: piwiPre synchronizes piwigo albums that were modified (enable-database)
# REQ 0007: piwiPre runs on windows and Linux development stations
# REQ 0008:  --dump-config folder allows to debug the configuration hierarchy (program_2)

# REQ 0009: piwiPre configures automatically album by ADDING piwiPre.ini files (autoconfiguration) (program_6, 7, 8, 9)
# REQ 0010: hand-writing a piwiPre.ini in albums may prevent any modification of files
# REQ 0011: during renaming, the name of .txt files is not changed. They simply go to the appropriate folder
# REQ 0012: tox is run on gitlab, generates html doc and coverage

# REQ 0020: 'enable-XXX' : false: feature is never active
# REQ 0021: 'enable-XXX' : true: feature is done in TRIAGE, and in ALBUMS if the result is not present


# REQ 0049: requirements are parsed from the source and put automatically in doc

# REQ 0050: temporary files are put in a real temporary dir
# REQ 0051: parser: remove trailing / at end of directories, this is common error

# REQ 0100: Renaming is based on the 'names' configuration template
# REQ 0101: Renaming takes into account the date/time of the picture shooting, found in metadata
# REQ 0102: if renaming a file clobber an existing file, piwiPre increments the last numerical field to avoid conflict


# REQ 0200: piwiPre verifies the albums are aligned with configuration (verify-album)
# REQ 0201: piwiPre realigns pictures rotation
# REQ 0202: piwiPre updates metadata
# REQ 0203: piwiPre generates lacking thumbnails from album
# REQ 0204: album modified pictures are saved in 'BACKUP'
# REQ 0205: piwiPre removes useless thumbnails
# REQ 0206: the albums to verify are specified by 'verify-album'
# REQ 0207: insert the 'author' metadata
# REQ 0208: use XMP metadata
# REQ 0209: --verify-album on all subdirectories when --recursive-verify-album is set, which is not the default
# REQ 0210: (BUG) if --enable-metadata false, thumbnails have no metadata, IPTCinfo.save() raises an exception  # noqa
# REQ 0211: (BUG) picture 1994-05-Avril-05-Thorigne-010.jpg, date should be found                               # noqa
#

# REQ 0212 :  trust-date-in-filename true/false
#   DOC:
#       If there is a date in the filename, (according to the 'names' argument), then this date is used for metadata
#       Else, if there is a date in metadata, this one is kept
#       Else, the file creation time is used, and written in metadata
#       So, if a file is checked twice, the 2nd run does not perform any change
#
# REQ 0213: --auto-config sets the directory where piwiPre.ini is read recursively: REJECTED

# REQ 0214: during verify-album, if the picture has been changed since last metadata update,
#    or if metadata in db and file are different, then the information in database is reset,
#    and md5 computed again

# REQ 0215: if --enable-metadata-reset true (default : false), metadata is reset to output of current configuration
#           if false, only empty metadata is filled
# REQ 0216: during verify-album, if a video is not inserted in the database, issue a warning
# REQ 0217: --stop-on-warning
# REQ 0218: --trace-malloc
# REQ 0219: --restart-from-dir

# REQ 0221: manage png and other image formats, convert to jpg if enable-rename true
# REQ 0222: manage avi and other video formats, convert to mp4 if enable-rename true
# REQ 0223: video conversion is done with ffmpeg
# REQ 0224: video tag extraction is done with ffprobe
# REQ 0225: ffmpeg-path
# REQ 0226: --piwigo-first-album
# REQ 0228: report the number of files modified, inserted, deleted in filesystem/database
# REQ 0229: special syntax verify-album "*" verifies all the sub albums of the root
# REQ 0230: --enable-conversion
# TODO REQ 0231: BUG: the Android app Piwigo NG hangs while playing video, while browsers play video OK on linux/windows
# REQ 0232:  --language [en, fr], chosen by default with the locale
# REQ 0233: doc in FRENCH
# REQ 0234: --enable-pwg-representative
# REQ 0235: --enable-colors
# REQ 0236: --reset-ini triage/remote/local/default
# REQ 0237: --language sets the default value of names and instructions
# REQ 0238: --auto-install
# REQ 0239: --piwiPre-path dir

# REQ 0300: piwiPre maintains (in album context) the database structure
# REQ 0301: piwiPre creates (in triage context) the database structure
# REQ 0302: piwiPre uses MD5 checksum to compare local and remote files
# REQ 0303: Compute MD5 checksums and add them directly in the database when doing a synchro
# REQ 0304: piwiPre reorders the piwigo directories with the right order !  (future piwiPre version 2.0)

# REQ 0306: generate an error if HOME/.piwiPre.ini is not protected: CANCELLED, because chmod is limited in windows
# REQ 0307: clean useless directories in WEB (useless files are removed while doing verify-album)
# REQ 0308: installer that allows to download piwiPre, ffmpeg and mariaDb
# REQ 0309: graphical UI for installer
# TODO REQ 0310 : installer: Check if MariaDB is installed before installing it
#    This seems non trivial: piwiPre works, for the python and exe version, without installing mariaDB
#         This has been tested on a PC that has no reasons to have mariaDB CC installed
#         That being said, maybe some program has installed maria DB...
#    The only reference to MariaDB in the registry is
#         HKEY_LOCAL_MACHINE\SOFTWARE\WOW6432Node\MariaDB Corporation\MariaDB Connector C
#    There is no MariaDB directory in "C:\Program Files (x86)" or "C:\Program Files"
#          or "C:\Users\fabien-local\AppData\Local"
#    Hypothesis: MariaDB CC is required on SOME database connections, not all.
#    Conservative solution: installer still allows to install MariaDB CC, just in case
# DONE: test enable-auto-configuration program 36, 37, 38, 436, 437,438

# -----------------------------------------------------------------------------------
# Requirements, Testing

# REQ 3000: autotest available with tox and pytest
# REQ 3001: piwipre> tests\\pwp_test.py -n number # runs the test number n                          # noqa
# REQ 3002: pwp_test -n 0 # runs all the tests
# REQ 3003: pwp_test can run all the autotests
# REQ 3004: PyCharm tests cases are saved in project files
# REQ 3005: all automatic tests run by pwp_test can be run by anyone on any development server
# REQ 3006: pwp_test assumes a valid configuration in HOME/.piwiPre.ini
# REQ 3007: coverage is run automatically by tox
# REQ 3008: coverage > 92 %
# REQ 3009: tests --dryrun, with triage and album
# REQ 3010: test error generation

# REQ 3100: test jpg files in TRIAGE, with/without rotation and metadata
# REQ 3101: test .txt files in TRIAGE
# REQ 3102: test .mp4 files in TRIAGE with metadata
# REQ 3104: tests end-2-end with piwogo sync (program_13)
# REQ 3105: test --version
# REQ 3106: test --chdir
# REQ 3107: test --licence
# REQ 3108: autotest of PwpPatcher
# REQ 3109: autotest of PwpPatcher should cover 80% of code
# REQ 3110: autotest of PwpMP4
# REQ 3111: There are no explicit remote locations on server, use config instead
# REQ 3112: someone with a valid config would run the same test on a different configuration
# REQ 3113: test --dump-config folder
# REQ 3114: test PwpError generation (program_26)
# REQ 3115: on argument error, raise an exception , not trapped
# REQ 3116: test the insertion of the 'author' metadata
# REQ 3117: BUG: program_11 generates a bug in the communication with the
#           SQL server, probably due to its own communication with git.
#           How to reproduce: run program_11 then program_400 in pycharm RUN mode.
#           if the same test programs are run in PyCharm DEBUG mode, no problem.
#           Explanation: stdin was closed, just reopen it!

# REQ 3120: test --enable-rename false  (program_27)
# REQ 3122: test dates (all cases) :  program_24
# REQ 3123: test unsupported (for instance .foo) files in TRIAGE
# REQ 3123: test pattern with {count} : program_36
# REQ 3124: test manual piwiPre.ini in album


# DONE: a modified copy of a picture (so same metadata!) is renamed thanks to collision avoiding

# DOC: piwigo user and password in HOME .piwiPre.ini
# Doc:  HOME/.piwiPre.ini contains only server-side information
# Doc: warning: .ini files should be written in UTF8 encoding !
# Doc: usage: photo is accessible or synchronized, not web: this is not an issue, piwigo will generate thumbnails

# REQ 4000: manage remote WEB, accessible with ssh
# REQ 4001: setup ssh/sftp session
# REQ 4002: We assume that thumbnails are always coming from the corresponding file in ALBUM
# REQ 4003: a thumbnail is created only if the file in ALBUM is new, or the thumbnail was non-existent
# REQ 4004: thumbnails are created in WEB, then copied to remote-web

# REQ 5001: piwiPre runs on a workstation connected to the piwigo server through ssh, without shared directories

# ----------------------------- FUTURE ROADMAP -------------------------------------
# Future V2.0
#
#
# TODO REQ 9000: piwiPre runs on a synology NAS
# TODO REQ 9001: --report-album: checks what should be done, prints only required changes, does nothing
# TODO REQ 9002: fully implement enable-conversion of images by NOT forcing working to be JPG, CAVEAT : testing...

# TODO REQ 9003: Parse requirements: reorder by REQ number, yell if duplicate number.
# TODO REQ 9100: piwiPre generates synology @eaDir thumbnails to speedup processing


class PwpMain:
    allowed_chars = r"a-zA-Z0-9\-_.&@~!,;+°()àâäéèêëïîôöùûüÿçñÀÂÄÉÈÊËÏÎÔÖÙÛÜŸÇÑ "    # noqa

    def __init__(self, arguments=None):
        self.initial_cwd = os.getcwd()
        self.parser = PwpParser(arguments=arguments, program="piwiPre", with_config=True)
        self.parser_config = self.parser.config
        self.dumped_config = None      # The config reached by --dump-config, used for test only
        self.dirs_to_synchronize = []
        self.expected_thumbnails = []  # list of expected thumbnails, when processing ALBUM. reset for each dir.
        self.expected_representatives = []  # list of expected representatives, when processing ALBUM, reset @ each dir

    @staticmethod
    def get_base_from_dir(dir_name, p_config):
        base = re.sub(r".jpg$", "", dir_name)
        base = re.sub(r'[^' + PwpMain.allowed_chars + ']', "_", base, re.IGNORECASE)

        # Maybe we are processing a directory which is already the result of a renaming process
        # so its format should be the same then the directory part of 'names'

        template = os.path.basename(os.path.dirname(p_config['names']))
        res = ACTOR.get_info_from_format(template, dir_name)
        if res and res['base'] is not None and res['base'] != '':
            return res['base']
        return base

    @staticmethod
    def build_path(base, common_path: str, path):
        result = base + '/' + common_path + ('' if common_path == '' else '/') + path
        if result[-1] == '/':
            result = result[:-1]
        return result

    def add_to_expected_thumbnails(self, filename: str, config: PwpConfig):
        # we need to get remote_web_filename BEFORE lower, because it relies
        # on comparing with directories
        remote_file = PwpJpg.get_remote_web_filename(config, filename)
        # we need to lower(), because files may have been created with a different case
        filename = filename.lower()
        remote_file = remote_file.lower() if remote_file else None
        if filename in self.expected_thumbnails:
            return

        self.expected_thumbnails.append(filename)
        self.expected_thumbnails.append(remote_file)

    def is_expected_thumbnail(self, filename: str):
        return filename.lower() in self.expected_thumbnails

    @staticmethod
    def clean_thumbnails(source_filename: str, config: PwpConfig, common_path: str):
        if config['enable-thumbnails'] is False:
            return  # pragma: no cover

        f = os.path.basename(source_filename)[:-4]  # skip .jpg

        thumbs_base = config['web'] + '/' + common_path
        thumbs_dir = PwpDirEntry.open(thumbs_base, config, context='web')

        thumbs = config['piwigo-thumbnails']

        for name, values in thumbs.items():
            thumb_name = name.format(f=f)
            filename = thumbs_base + '/' + thumb_name
            if filename in thumbs_dir.files:
                thumbs_dir.files[filename].remove(father=thumbs_dir, msg="removing old thumbnail")

    def verify_thumbnails(self, source: PwpJpg, config: PwpConfig, common_path: str, stage: str,
                          force_rebuild: bool,
                          summary: PwpSummary):

        # force_rebuild: if true, must rebuild the thumbnail because file HAS changed

        source_filename = source.filename
        if config['enable-thumbnails'] is False:   # pragma: no cover
            LOGGER.debug(f"{stage} {source_filename} enable-thumbnails is False")
            return

        f = os.path.basename(source_filename)[:-4]  # skip .jpg

        thumbs_base = config['web'] + '/' + common_path
        thumbs_dir = PwpDirEntry.open(thumbs_base, config, context='web')
        thumbs_dir.read()

        thumbs = config['piwigo-thumbnails']

        for name, values in thumbs.items():
            width = values['width']
            height = values['height']
            crop = values['crop']
            thumb_name = name.format(f=f)

            local_name = thumbs_base + '/' + thumb_name
            self.add_to_expected_thumbnails(local_name, config)

            if config['dryrun']:
                # CAVEAT: conflict management has NOT occurred, because the files are not created in ALBUM
                #         So we do *not* know the exact thumbnail name
                #         hence the message is not accurate
                LOGGER.info(f"Would create Thumbnail {width}x{height} crop={crop} for {source_filename}")
            else:
                source.thumbnail(local_name, width, height, crop, config, force_rebuild, thumbs_dir, summary)

        index_base = 'index.htm'
        local_name = thumbs_base + '/' + index_base
        pwp_index = PwpFileEntry.lookup(local_name, config=config, context='web')
        self.add_to_expected_thumbnails(local_name, config)

        if pwp_index.is_local:
            pass  # index is already in local or remote, or we do not need to create
        else:
            if config['dryrun']:
                LOGGER.info(f"would create thumbnail html index '{local_name}'")
            else:
                LOGGER.info(f"Create thumbnail html index '{local_name}'")
                ACTOR.mkdirs(os.path.dirname(local_name))
                with ACTOR.open(local_name, 'w') as f:
                    print("Not allowed!", file=f)
                pwp_index.is_local = True
                summary.thumb_index = True

        if pwp_index.put():
            summary.rem_thumb_index = True

    # ------------------------ management of database

    def check_representative(self, source: PwpData, target, config: PwpConfig, summary: PwpSummary):
        """
        if this is a VIDEO, creates the representative picture
        :param source: db description of the source image
        :param target: the path of the file to be inserted
        :param config: current configuration
        :param summary:
        :return: None
        """
        if source.representative is None:
            return
        if not config["enable-pwg-representative"]:
            return  # pragma: no cover

        # so, we have a video
        self.expected_representatives.append(source.representative)     # it is expected, even if it is already here
        rep_dir = os.path.dirname(source.representative)
        ACTOR.mkdirs(rep_dir)

        if ACTOR.isfile(source.representative):
            rep_date = os.path.getmtime(source.representative)
            file_date = os.path.getmtime(target)
            if rep_date >= file_date:
                LOGGER.info(f"Representative {source.representative} is more recent than {target}")
                return
            ACTOR.delete(source.representative)

        # build the representative
        PwpVideo.build_representative(target, source.representative, config=config,
                                      new_author=source.author,
                                      new_copyright=source.copyright,
                                      new_special=source.special,
                                      new_date=source.creation)
        representative_entry = PwpFileEntry.lookup(source.representative, config=config, context='album')
        summary.representative = True
        representative_entry.put()

        source.incr_logger_count("Video representative")

    def insert_in_database(self, source: PwpData, target, sql_filename, config: PwpConfig, summary: PwpSummary):
        """
        insert sql_filename in the sql database.
        if this is a VIDEO, also creates the representative picture
        :param source: db description of the source image
        :param target: the path of the file to be inserted
        :param sql_filename: the path of the file, relative to album
        :param config: current configuration
        :param summary: logger for actions taken
        :return: sql_get_file_info()
        """

        file_info = ACTOR.sql_insert_file(target, sql_filename)
        self.check_representative(source, target, config, summary)

        return file_info

    def verify_database(self, stage, config, source: PwpData, summary: PwpSummary):
        """
        if necessary, updates the JPG metadata of the file inside the DB
        :param stage: current stage
        :param config: current config
        :param source: db description of the source image
        :param summary: summary of actions done
        :return description of what has been updated, or ' db[] ' if no update
        """
        target = source.filename

        self.check_representative(source, target, config, summary)

        if not config['enable-database']:  # pragma: no cover
            LOGGER.debug(f"{stage} {target} enable-database is False")
            return

        if not source.can_be_inserted_in_db:  # pragma: no cover
            LOGGER.debug(f"{stage} {target} is not supported by piwigo: no insertion in database")
            return

        is_modified = False

        sql_filename = target.replace(config['album'], '')
        # example of sql_filename: '/1988/1988-07-Juillet-21-Mariage/1988-03-Mars-15-Rennes-001.jpg'  # noqa

        src_width = source.width
        src_height = source.height

        src_size = int(source.size / 1024)

        sql_file_info: FileInfo = ACTOR.sql_get_file_info(sql_filename)

        if sql_file_info is None:
            # this file is not in the sql database
            sql_file_info = self.insert_in_database(source, target, sql_filename, config, summary)
            summary.db_created = True
            summary.db_size = True
            summary.db_width = True
            summary.db_height = True
            summary.db_md5 = True
            src_md5 = sql_file_info.md5sum
            is_modified = False
        else:
            if sql_file_info.file_size != src_size:
                summary.db_size = True
                is_modified = True

            if sql_file_info.width != src_width:
                summary.db_width = True
                is_modified = True

            if sql_file_info.height != src_height:
                summary.db_height = True
                is_modified = True

            # CAVEAT: lastmodified is the last modification of the database, not of the file            # noqa
            # So, it is under responsibility of programmer to ensure that the data is kept OK
            # NB: changing the md5 also updates lastmodified.                                           # noqa

            # be paranoid: ALWAYS compute MD5 and check
            src_md5 = ACTOR.compute_md5(target)

            if sql_file_info.md5sum != src_md5:
                summary.db_md5 = True
                is_modified = True

        if ((sql_file_info.latitude != source.latitude and source.latitude) or
                (sql_file_info.longitude != source.longitude) and source.longitude):
            summary.db_gps = True
            is_modified = True

        src_author = source.author if source.author else 'NULL'
        if sql_file_info.author != src_author:
            summary.db_author = True
            is_modified = True

        if not is_modified:
            return

        # the file has been modified, let's recompute them and update all data

        source.incr_logger_count("Database")
        ACTOR.sql_set_data(sql_filename, src_md5, src_size, src_width, src_height,
                           source.latitude, source.longitude, source.author,
                           warn_if_no_change=True)
        return

    @staticmethod
    def same_file(f1: str, f2: PwpFileEntry, config: PwpConfig):
        f1_entry = PwpFileEntry.lookup(f1, "local", config)
        if not f1_entry.md5sum:
            f1_entry.md5sum = ACTOR.compute_md5(f1_entry.local)

        if f2.is_local:
            if not f2.md5sum:
                f2.md5sum = ACTOR.compute_md5(f2.local)
            return f1_entry.md5sum == f2.md5sum

        f2_sum = f2.get_remote_md5()
        return f1_entry.md5sum == f2_sum

    @staticmethod
    def rename_allowed(_stage, config):
        if config['enable-rename'] is False:
            return False
        return True

    @staticmethod
    def build_auto_config(src, dst, base, config: PwpConfig):
        # we want to change the {base} component of names to base
        # because this value cannot be guessed from the directory name in ALBUM
        with ACTOR.open(src, 'r') as s:
            lines = s.readlines()
        ACTOR.mkdirs(os.path.dirname(dst), forced=True)
        with ACTOR.open(dst, 'w') as d:
            d.write(f"# file generated by --enable-auto-configuration on {datetime.datetime.now()}\n")
            for li in lines:
                m = re.match(r"names\s*:(.*)", li)
                if m:
                    li = li.replace('{base}', base)
                d.write(li)
        auto_entry = PwpFileEntry.lookup(dst, config=config, context="web")
        auto_entry.put()

    #     when enable-remote-album is True and remote-album set (and ssh information valid)
    #     album is used as a cache to remote-album
    #     its value SHOULD be set a *local* directory, typically this is 'ALBUM'
    #     after processing triage or album, album and remote-album are coherent (for the processed directories)
    #     so that the user can look inside album to see the state of remote-album
    #     if a file is in album but not in remote-album,
    #     it is considered as abnormal and removed

    def manage_conflict(self, common_path: str, old_file_path: str, new_date: str, new_author: str, base: str,
                        current_file: PwpData, config, stage, summary: PwpSummary):
        """

        :param common_path:
        :param old_file_path:  old_file.local
        :param new_date:
        :param new_author:
        :param base:
        :param current_file: old_file or a copy with metadata and rotation
        :param config:
        :param stage:
        :param summary:
        :return: target, move_to_album, new_filepath
        """
        move_to_album = (stage == 'triage')
        # if True, we need to move the file to album
        # in album stage, we move again to album only if the file has been modified

        if current_file.is_modified:
            move_to_album = True
            if not config['dryrun']:
                # this may happen if the picture was rotated but no modification of metadata
                current_file.save_with_info(config)

        current_file.close()

        file_format = config.format('names')
        file_dico = config.format_dict(new_date, new_author, base=base, suffix=current_file.suffix)

        if self.rename_allowed(stage, config):
            new_filepath = file_format.format(**file_dico)
        else:
            # we do not change the name
            new_filepath = common_path + '/' + os.path.basename(old_file_path)

        if config['enable-conversion']:
            new_filepath = current_file.patch_after_rename(new_filepath)

        # CAVEAT: in ALBUM, the path between ALBUM and the picture is reset by file_format !!!
        # so, common_path is NOT part of new_filepath !

        target = PwpFileEntry.lookup(config['album'] + '/' + new_filepath, context="album", config=config)
        # lookup creates father, the enclosing directory

        to_increment = ACTOR.get_last_numerical_field(config['names'])

        if not target.is_local:
            move_to_album = True  # it is not here, we need to put it

        # if os.path.abspath(target.local) != os.path.abspath(current_file.filename):
        if target.local != current_file.filename:
            move_to_album = True

        summary.action = " Copy"
        while target.exists():  # one file is present in the target directory, with the same name. remote or local

            # first, we clean the relationship between remote and local
            if target.is_local:
                target.local_coherent_with_remote()

            if self.same_file(current_file.filename, target, config):                        # case 1) of conflicts.rst
                if target.local == old_file_path:
                    LOGGER.info(f"File '{old_file_path}' has not changed")
                else:
                    LOGGER.info(f"New file '{old_file_path}' is already in album as '{target.local}'")
                summary.action = " Keep"
                move_to_album = False
                break

            elif target.local == old_file_path:                                              # case 2) of conflicts.rst
                # current_file is a modification of old_file_path with metadata etc. So we need to update old_file_path
                LOGGER.debug(f"Update '{old_file_path}' due to modifications")
                summary.action = "Updat"    # noqa
                move_to_album = True
                break
            elif self.rename_allowed(stage, config):                                         # case 3) of conflicts.rst
                file_dico[to_increment] += 1
                new_filepath = file_format.format(**file_dico)
                if config['enable-conversion']:
                    new_filepath = current_file.patch_after_rename(new_filepath)
                target = PwpFileEntry.lookup(config['album'] + '/' + new_filepath, context="album", config=config)
                move_to_album = True
                summary.action = "Renam"    # noqa
            else:                                                                            # case 4) of conflicts.rst
                LOGGER.debug(f"Clobber '{old_file_path}' because rename not allowed")
                move_to_album = True
                summary.action = "Clobb"    # noqa
                break

        summary.destination = target
        return target, move_to_album, new_filepath

    def run_stage_file(self, stage: str, config: PwpConfig, common_path: str,
                       old_file: PwpFileEntry, base: str):

        LOGGER.debug('')

        summary = PwpSummary(stage, old_file)  # will hold a 1 line summary of actions taken

        LOGGER.debug(f"{stage} file path='{common_path}' filename='{old_file.basename}' base='{base}'")

        backup = self.build_path(config['backup'], common_path, old_file.basename)
        copy_file_path = self.build_path(config['tmp_dir'], common_path, old_file.basename)

        if not old_file.local_coherent_with_remote():
            # be silent, local_coherent_with_remote has already generated a warning
            ACTOR.copy(old_file.local, backup)
            ACTOR.delete(old_file.local)
            # we delete the local version, but will process the remote one
            if not old_file.is_remote:
                # no local and no remote...
                summary.action = "DELET"        # noqa
                summary.backup = backup
                LOGGER.msg(summary)
                return

        if not old_file.get():
            # LOGGER.msg("Unable to get file from remote location, abort")
            # be silent, get has already generated a warning
            summary.action = "ABORT"
            LOGGER.msg(summary)
            return

        current_file = PwpData.create(old_file.local, config=config, stage=stage,
                                      tmp=copy_file_path, backup=backup)

        current_file.incr_logger_count()
        if current_file.to_abort:  # i.e. PwpAvoided
            # be silent on this type of file
            return

        new_date, new_author = current_file.compute_new_date_author(config)

        summary.author = new_author
        summary.date = PwpData.date_to_str(new_date)

        current_file, _mod = current_file.verify_orientation(stage, config, summary)

        allow_reset = config['enable-metadata-reset']

        # -----------------------------------------------------------------------------------------
        current_file, _mod = current_file.verify_metadata(stage, config, summary,
                                                          new_date=new_date, new_author=new_author,
                                                          enable_reset=allow_reset)
        # -----------------------------------------------------------------------------------------

        # CAVEAT: verify_orientation and verify_metadata MAY change current_file
        #         it MAY be a modified copy in tmp_dir, with new metadata and new rotation
        #

        # -------------------------------------------------------------------------------------------------
        target, move_to_album, new_filepath = self.manage_conflict(common_path, old_file.local,
                                                                   new_date, new_author, base, current_file,
                                                                   config, stage, summary)
        # --------------------------------------------------------------------------------------------------

        target_rel_path = os.path.dirname(new_filepath)

        if move_to_album:

            if stage == 'album':
                LOGGER.info(f"backup {old_file.basename} to {backup}")
                ACTOR.copy(old_file.local, backup)
                summary.backup = backup
                current_file.incr_logger_count("Backup")

                if current_file.filename != old_file.local:
                    # the original file has been first copied to working,
                    # and then we will copy from working to target
                    # so, we can remove file_path.
                    # if enable-rename is false, target will be the same file thant file_path, so removing is OK
                    # is enable-rename is true, then maybe target is a different file, and we should remove file_path
                    # see program_41 and sample-mov.MOV, which SHOULD be removed.
                    ACTOR.delete(old_file.local)  # if dryrun, does nothing :-)

            if config['dryrun']:
                if old_file.local == target.local:
                    LOGGER.info(f"Would update '{old_file.local}'")
                else:
                    LOGGER.info(f"Would rename '{old_file.local}' : '{target.local}'")
                    # otherwise message is misleading: in reality, we do not rename to itself,
                    # we rename the copy that has been changed
            else:
                if current_file.filename == target.local:
                    raise PwpInternalError(f" '{current_file}' == '{target.local}'")

                self.clean_thumbnails(source_filename=target.local, config=config, common_path=common_path)
                ACTOR.move(current_file.filename, target.local)
                target.is_local = True
                if target.put():
                    summary.remote = True

                current_file.incr_logger_count("Renamed")
                if old_file.local == target.local:
                    LOGGER.info(f"Update '{old_file.local}'")
                else:
                    LOGGER.info(f"RENAME: '{old_file.local}' : '{target.local}'")
        else:
            # i.e. move_to_album False : target is the same file as local or remote,
            # if remote and local exist simultaneously, they have the same md5 (because local_coherent_with_remote)
            # but the file may be absent from local or from remote
            target.synchronize()

        # here, target is always in ALBUM

        # even if the picture is unchanged, maybe the thumbs are not done

        if config['dryrun']:
            target_object = PwpData.create(old_file.local, config=config, stage=stage)
            # We have to cheat, because target is NOT build, so we use file_path !
        else:
            target_object = PwpData.create(target.local, config=config, stage=stage)

        # --------------------------------------------------------------------------------------------
        self.verify_database(stage, config, target_object, summary)
        # --------------------------------------------------------------------------------------------

        # --------------------------------------------------------------------------------------------
        self.verify_thumbnails(target_object, config, target_rel_path, stage, move_to_album, summary)
        # --------------------------------------------------------------------------------------------

        if target_object.representative:
            rep_file = PwpData.create(target_object.representative, config=config, stage=stage)
            self.verify_thumbnails(rep_file, config,
                                   target_rel_path + '/pwg_representative',
                                   stage, move_to_album, summary)

        target_object.close()

        if not move_to_album and current_file.filename != old_file.local:
            ACTOR.delete(current_file.filename, forced=True)

        if stage != 'triage':
            LOGGER.msg(summary)
            return

        if not config['enable-auto-configuration']:  # pragma: no cover
            LOGGER.msg(summary)
            return

        this_ini = self.build_path(config[stage], common_path, config['ini-file'])
        config_path = os.path.dirname(new_filepath)
        auto_config_ini = self.build_path(config['web'], config_path, config['ini-file'])

        if ACTOR.isfile(this_ini) and not ACTOR.isfile(auto_config_ini):
            if config['dryrun']:
                LOGGER.info(f"Would Auto-configure '{this_ini}' to '{auto_config_ini}'")
            else:
                LOGGER.info(f"Auto-configure '{this_ini}' to '{auto_config_ini}'")
                self.build_auto_config(this_ini, auto_config_ini, base, config)
                summary.auto_conf = True
                # if config['enable-remote-web'] and config['remote-web']:
                #     remote_ini = auto_config_ini.replace(config['web'], config['remote-web'])
                #     ACTOR.remote_put(this_ini, os.path.dirname(remote_ini))   # noqa
                #     summary.rem_auto_conf = True

        LOGGER.msg(summary)

    def remove_useless_thumbnails(self, path, config):
        # test me using program_402
        # if config['dryrun']:
        #     LOGGER.debug("dryrun is True, abort remove_useless_thumbnails")
        #     return
        # if we abort here, many tests will fail
        if not config['enable-thumbnails']:  # pragma: no cover
            LOGGER.debug(f"{path} enable-thumbnails is False, abort remove_useless_thumbnails")
            return
        if not config['enable-thumbnails-delete']:
            LOGGER.debug(f"{path} enable-thumbnails-delete is False, abort remove_useless_thumbnails")
            return

        # We can always erase both local and remote files,
        # because config[web] CANNOT be the mount point to remote, it must be a temporary folder, e.g. WEB.

        # CAVEAT: remove_useless_thumbnails must NOT be recursive with subdirectories,
        # because this is already taken account by run_stage_dir

        thumbs_base = config['web'] + '/' + path
        father = PwpDirEntry.open(thumbs_base, config, 'web')

        for file in father.all_entries():
            if self.is_expected_thumbnail(file.local or file.remote):
                pass  # this is normal
            elif file.basename[-4:] == '.ini':
                pass
            else:
                file.remove(father, "when removing extra thumbnail")

        if father.is_empty():
            father.remove(None, "when removing empty thumbnail directory")

        self.expected_thumbnails = []
        return  # because the debugger sometimes has issues with tail return

    # -----------------------------------------------------------------------------------------
    # SQL Database management

    @staticmethod
    def verify_sql_dir(p_config: PwpConfig, path: str):
        if not p_config["enable-database"]:
            return True

        real_path = p_config["album"] + '/' + path

        all_entries = os.listdir(real_path) if os.path.isdir(real_path) else []
        files_list = []
        dirs_list = []
        for item in all_entries:
            if ACTOR.isfile(real_path + '/' + item):
                files_list.append(item)
            else:
                dirs_list.append(item)
        files_list.sort()
        dirs_list.sort()

        dir_info: DirInfo = ACTOR.sql_get_dir_info(path)
        sql_file_descr: dict[FileInfo] = ACTOR.sql_get_dir_file_list(dir_info)

        for file in files_list:
            if file not in sql_file_descr:
                if PwpData.get_type(file) in (PwpData.JPG, PwpData.MP4, PwpData.IMAGE, PwpData.VIDEO):
                    # NB: In ALBUM case, the insertion of lacking files is done BEFORE,
                    # so we should never end up here
                    raise PwpError(f"SQL ERROR: database descriptor for file '{file}' " +
                                   f"should have been inserted in '{path}'")
                else:
                    pass
                    # txt etc files are not managed

        for file in sql_file_descr.keys():
            if file not in files_list:
                # LOGGER.warning(f"database descriptor for file '{file}' should be deleted from '{path}'")
                ACTOR.sql_remove_file_from_db(p_config, sql_file_descr[file])

        sql_sons_descr = ACTOR.sql_get_dir_sub_dirs(dir_info)

        rank = 1
        for subdir in dirs_list:

            if subdir not in sql_sons_descr.keys():
                # LOGGER.warning(f"db for dir '{subdir}' should be inserted in '{path}' @ {index}")
                ACTOR.sql_insert_dir_at_rank(subdir, dir_info, rank)
            elif rank != sql_sons_descr[subdir].rank:
                # LOGGER.warning(f"rank of dir '{subdir}' in '{path}' {sql_sons_descr[subdir].rank} -> {index}")
                ACTOR.sql_change_dir_rank(dir_info, sql_sons_descr[subdir], rank)
            rank += 1

        for subdir in sql_sons_descr:

            if subdir not in dirs_list:
                # LOGGER.warning(f"database descriptor for dir '{subdir}' should be deleted from '{path}'")
                ACTOR.sql_remove_dir_from_database_recursive(p_config, sql_sons_descr[subdir])

    @staticmethod
    def build_representative_config(config: PwpConfig):
        """
        Creates a new configuration based on the current one, dedicated to processing the representative files
        :param config: current config, MAYBE in triage mode
        :return: new config, in album mode
        """
        new = PwpConfig(filename="Fake for representative",
                        dico={'enable-rename': 'false',
                              'enable-metadata': 'true',
                              'enable-thumbnails': 'true',
                              'triage': None,
                              # verify-album is useless
                              # enable-metadata: keep the current value
                              # enable-rotation: useless, we have never a rotation
                              # enable-database: keep the current value
                              # enable-remote-web: keep the current value
                              # recursive-verify-album: useless
                              # verify-album: useless
                              # enable-thumbnails-delete: keep the current value
                              })
        new.merge_ini(config)
        return new

    def clean_representatives(self, path):
        if not os.path.isdir(path):
            return
        rep_files = os.listdir(path)
        for file in rep_files:
            file_path = f"{path}/{file}"
            if file_path not in self.expected_representatives:
                LOGGER.msg(f"representative {file_path} is useless: deleted")
                ACTOR.delete(file_path)

    def verify_representatives(self, path, config: PwpConfig):
        if not os.path.isdir(path):
            return
        # If there are some pictures that require a representative, then the directory IS created

        if not path.startswith(config['album']):
            raise PwpInternalError(f"illegal path {path}")

        # le = len(config['album']) + 1  # + 1 to remove /
        # common_path = path[le:]
        self.run_stage_dir("album", config,
                           current_dir=PwpDirEntry.open(path, config=config, context="album"),
                           base="")
        # base can be "", because we use it only for rename

        # TODO: remove rep_dir if empty

    @staticmethod
    def get_common_path(path, stage, config: PwpConfig):
        # common path is the part of path starting just after TRIAGE or ALBUM, without leading /
        root = config[stage]
        if path == root:
            return ''
        if path.startswith(root):
            return path[len(root) + 1:]
        else:
            raise PwpInternalError(f"Illegal path {path} should start with {root}")

    # -----------------------------------------------------------------------------------------
    # run_stage_sir

    def run_stage_dir(self, stage: str, p_config: PwpConfig, current_dir: PwpDirEntry, base: str, recursive=True):
        """
        :param stage: triage,  album
        :param p_config: the configuration **when we entered the directory, before reading the local piwiPre.ini **
        :param current_dir: dir under processing
        :param base: the part of renamed files inherited from common_path
        :param recursive: recursively enter subdirectories
        """

        if re.match(r'(.picasaoriginals.*|.idea.*|@eaDir|.comments)', current_dir.basename):
            # nothing to do here, this directory should not be managed
            return   # pragma: no cover

        if stage == 'triage' and p_config['triage'] is None:
            raise PwpConfigError("No TRIAGE directory")

        if stage == 'album' and p_config['album'] is None:
            raise PwpConfigError("No ALBUM directory")

        # common path is the part of path starting just after TRIAGE or ALBUM
        common_path = self.get_common_path(current_dir.local, stage, p_config)

        cur_basename = current_dir.basename

        if stage == "triage":
            config_path = p_config["triage"] + ('' if common_path == '' else '/' + common_path)
        else:
            config_path = p_config['web'] + ('' if common_path == '' else '/' + common_path)

        new_conf_file = config_path + '/' + os.path.basename(p_config['ini-file'])
        new_conf = p_config.push_local_ini(new_conf_file)

        LOGGER.msg('')
        LOGGER.msg(f"------ {stage} dir: common_path='{common_path}' base='{base}'")

        self.expected_thumbnails = []

        new_base = self.get_base_from_dir(cur_basename, new_conf)

        if new_base:
            base = new_base

        all_files, all_dirs = current_dir.read()

        if recursive:
            restart = new_conf['restart-from-dir']
            restart_dir = os.path.basename(restart) if restart and os.path.dirname(restart) == common_path else None

            for item in all_dirs:
                if item.basename == restart_dir:
                    restart_dir = None
                if restart_dir:
                    # we do not manage subdirectories until restart_dir has been seen
                    LOGGER.msg(f"--restart-from-dir '{restart}' excluded '{item}'")
                    continue
                if item.basename == "pwg_representative":
                    continue  # because this directory is special and managed elsewhere
                self.run_stage_dir(stage, new_conf, item, base)

            if restart_dir:
                raise PwpError(f"--restart-from-dir '{restart}' unseen directory: abort")

        self.expected_thumbnails = []
        self.expected_representatives = []
        # if recursive, run_stage_dir has already managed its own remove_useless_thumbnails

        # snapshot1 = ACTOR.trace_malloc_snapshot("start run_stage_file")

        for item in all_files:
            self.run_stage_file(stage, new_conf, common_path, item, base)
            # this should populate expected_thumbnails, even if building it is not necessary

        for item in all_dirs:
            web_path = new_conf['web'] + ('' if common_path == '' else '/' + common_path)
            self.add_to_expected_thumbnails(web_path + '/' + item.basename, p_config)
            # prev line is a hack to ensure simply that subdirectories are kept

        if stage == 'album':
            self.remove_useless_thumbnails(common_path, new_conf)
            # must be done BEFORE representative management,
            # because  remove_useless_thumbnails uses self.expected_representatives
            # which is reset by verify_representatives

        if new_conf['enable-database']:
            rep_dir_set = set([os.path.dirname(f) for f in self.expected_representatives])
            rep_config = self.build_representative_config(new_conf)
            for rep_dir in rep_dir_set:
                self.clean_representatives(rep_dir)  # do cleaning first, because verify will delete expected_rep
            for rep_dir in rep_dir_set:
                self.verify_representatives(rep_dir, rep_config)

        if stage == 'album':
            self.verify_sql_dir(new_conf, common_path)
            # self.remove_useless_thumbnails(common_path, new_conf)

        PwpDirEntry.clean_cache()

        # ACTOR.trace_malloc_snapshot("end run_stage_file", snapshot1, garbage=True)

        # PwpObject.check_leaks()

    # run_stage_sir
    # -----------------------------------------------------------------------------------------

    def combined_config(self, path: str, p_config, caller="dump_config"):
        """
        Goes through the hierarchy of ALBUM, up to path included,
        and merges with the corresponding configuration in WEB

        :param path: MUST be a relative path, the code here does NOT manage absolute paths
        :param p_config: current config
        :param caller: string for debug
        :return: the merged configuration
        """

        if p_config["triage"] and path.startswith(p_config["triage"]):
            cur_path = p_config["triage"]
            new_conf = p_config
            new_conf_file = cur_path + '/' + os.path.basename(p_config['ini-file'])
            new_conf = new_conf.push_local_ini(new_conf_file)

            length = len(p_config["triage"]) + 1  # +1 to get the following /
            rel_path = path[length:]
            if rel_path:
                dirs = rel_path.split('/')
                for cur_dir in dirs:
                    cur_path += '/' + cur_dir
                    if not os.path.isdir(cur_path):
                        raise PwpConfigError(
                            f"argument of {caller} must exist, and must be a subdir of triage", path)
                    new_conf_file = cur_path + '/' + os.path.basename(p_config['ini-file'])
                    new_conf = new_conf.push_local_ini(new_conf_file)
            self.dumped_config = new_conf  # just for test
            return new_conf

        if p_config["album"] and p_config["web"] and path.startswith(p_config["album"]):
            cur_path_config = p_config["web"]
            cur_album_path = p_config["album"]
            new_conf = p_config
            new_conf_file = cur_path_config + '/' + os.path.basename(p_config['ini-file'])
            new_conf = new_conf.push_local_ini(new_conf_file)

            length = len(p_config["album"]) + 1  # +1 to get the following /
            rel_path = path[length:]
            if rel_path:
                dirs = rel_path.split('/')
                for cur_dir in dirs:
                    cur_path_config += '/' + cur_dir
                    cur_album_path += '/' + cur_dir
                    if not os.path.isdir(cur_album_path):
                        raise PwpConfigError(
                            f"argument of {caller} must exist, and must be a subdir of album", path)
                    new_conf_file = cur_path_config + '/' + os.path.basename(p_config['ini-file'])
                    new_conf = new_conf.push_local_ini(new_conf_file)
            self.dumped_config = new_conf  # just for test
            return new_conf

        raise PwpConfigError(f"--{caller} : dir not existing in triage or album", path)

    def dump_config(self, path, p_config):
        conf = self.combined_config(path, p_config, caller="dump_config")
        LOGGER.info(f"dump_config({path}): {conf['ini-filename-parsed']}")
        pprint.pprint(conf)

    def reset_ini(self, arg):
        ini_file = self.parser_config['ini-file']
        ACTOR.backup(ini_file)
        if arg == 'triage':
            with ACTOR.open(ini_file, 'w') as f:
                f.write("# configuration file automatically generated by piwiPre --reset-ini triage\n")
                f.write("# Edit to setup your own values\n")
                f.write("\n")
                f.write("triage: TRIAGE         # where original pictures are read by piwiPre\n")
                f.write("album: ALBUM           # where processed pictures are written by piwiPre\n")
                f.write("backup: BACKUP         # where original pictures are back-up-ed by piwiPre\n")
                f.write("\n")
                f.write("web:                   # keep empty: no thumbnails\n")
                f.write("remote-web:            # keep empty: no remote directory\n")
                f.write("remote-album:          # keep empty: no remote directory\n")
                f.write("remote-user:           # keep empty: no remote directory\n")
                f.write("remote-pwd:            # keep empty: no remote directory\n")
                f.write("remote-host:           # keep empty: no remote directory\n")
                f.write("remote-port:           # keep empty: no remote directory\n")
                f.write("remote-incoming:       # keep empty: no remote directory\n")
                f.write("piwigo-user:           # keep empty: no remote directory\n")
                f.write("sql-user:              # keep empty: no sql database\n")
                f.write("sql-pwd:               # keep empty: no sql database\n")
                f.write("sql-host:              # keep empty: no sql database\n")
                f.write("sql-port:              # keep empty: no sql database\n")
                f.write("sql-database:          # keep empty: no sql database\n")
                f.write("\n")
                f.write("enable-rename: true               # This value is usually OK\n")
                f.write("enable-rotation: true             # This value is usually OK\n")
                f.write("enable-metadata: true             # This value is usually OK\n")
                f.write("enable-conversion: true           # This value is usually OK\n")
                f.write("trust-date-in-filename: true      # This value is usually OK\n")
                f.write("\n")
                f.write("enable-thumbnails: false          # keep false: no thumbnails\n")
                f.write("enable-thumbnails-delete: false   # keep false: no thumbnails\n")
                f.write("enable-remote-web: false          # keep false: no remote directory\n")
                f.write("enable-remote-album: false        # keep false: no remote directory\n")
                f.write("enable-database: false            # keep false: no remote directory\n")
                f.write("enable-auto-configuration: false  # keep false: no remote directory\n")
                f.write("enable-pwg-representative: false  # keep false: no remote directory\n")
                f.write("\n")
                f.write("# End of file\n")
            LOGGER.msg(f"Generated '{ini_file}' for a local usage\n")
        elif arg == 'default':
            self.parser.build_ini_file(ini_file)
            LOGGER.msg(f"Reset '{ini_file}' to default\n")
        elif arg == "local":
            with ACTOR.open(ini_file, 'w') as f:
                f.write("# configuration file automatically generated by piwiPre --reset-ini local\n")
                f.write("# Edit to setup your own values\n")
                f.write("\n")
                f.write("triage: TRIAGE       # where original pictures are read by piwiPre, you can keep this value\n")
                f.write("album: ALBUM         # CHANGE THIS to the mount-point of the piwigo album\n")
                f.write("backup: BACKUP       # where original pictures are copied by piwiPre, you can keep this\n")
                f.write("web: WEB             # CHANGE THIS to the mount-point of the piwigo thumbnails\n")
                f.write("\n")
                f.write("remote-web:          # keep empty: no remote directory\n")
                f.write("remote-album:        # keep empty: no remote directory\n")
                f.write("remote-user:         # keep empty: no remote directory\n")
                f.write("remote-pwd:          # keep empty: no remote directory\n")
                f.write("remote-host:         # keep empty: no remote directory\n")
                f.write("remote-port:         # keep empty: no remote directory\n")
                f.write("remote-incoming:     # keep empty: no remote directory\n")
                f.write("\n")
                f.write("# piwigo-user:         # in HOME/.piwiPre.ini: CHANGE THIS to the name of the piwigo user\n")
                f.write("# sql-user:            # in HOME/.piwiPre.ini: CHANGE THIS to the name of the sql user\n")
                f.write("# sql-pwd:             # in HOME/.piwiPre.ini: CHANGE THIS to the pwd of sql user\n")
                f.write("# sql-host:            # in HOME/.piwiPre.ini: CHANGE THIS to the sql host\n")
                f.write("# sql-port:            # in HOME/.piwiPre.ini: CHANGE THIS to the sql port\n")
                f.write("# sql-database:        # in HOME/.piwiPre.ini: CHANGE THIS to the sql database\n")
                f.write("\n")
                f.write("enable-rename: true               # This value is usually OK\n")
                f.write("enable-rotation: true             # This value is usually OK\n")
                f.write("enable-metadata: true             # This value is usually OK\n")
                f.write("enable-conversion: true           # This value is usually OK\n")
                f.write("trust-date-in-filename: true      # This value is usually OK\n")
                f.write("\n")
                f.write("enable-thumbnails: true           # keep true\n")
                f.write("enable-thumbnails-delete: true    # keep true\n")
                f.write("enable-remote-web: false          # keep false: no remote directory\n")
                f.write("enable-remote-album: false        # keep false: no remote directory\n")
                f.write("enable-database: true             # keep true\n")
                f.write("enable-auto-configuration: true   # This value is usually OK\n")
                f.write("enable-pwg-representative: true   # keep true\n")
                f.write("\n")
                f.write("# End of file\n")
            LOGGER.msg(f"Generated '{ini_file}' for a local usage")
            LOGGER.msg("")
            LOGGER.msg(f"your HOME is '{os.path.expanduser('~')}'")
            LOGGER.msg("")
            LOGGER.msg("The following value SHOULD be set in HOME/.piwiPre.ini:")
            LOGGER.msg("   piwigo-user:         # CHANGE THIS to the name of the piwigo user")
            LOGGER.msg("   sql-user:            # CHANGE THIS to the name of the sql user")
            LOGGER.msg("   sql-pwd:             # CHANGE THIS to the pwd of sql user")
            LOGGER.msg("   sql-host:            # CHANGE THIS to the sql host")
            LOGGER.msg("   sql-port:            # CHANGE THIS to the sql port")
            LOGGER.msg("   sql-database:        # CHANGE THIS to the sql database")
            LOGGER.msg("")
        elif arg == "remote":
            with ACTOR.open(ini_file, 'w') as f:
                f.write("# configuration file automatically generated by piwiPre --reset-ini remote\n")
                f.write("# Edit to setup your own values\n")
                f.write("\n")
                f.write("triage: TRIAGE       # where original pictures are read by piwiPre, you can keep this value\n")
                f.write("album: ALBUM         # temporary directory where processed pictures are written by piwiPre\n")
                f.write("backup: BACKUP       # where original pictures are copied by piwiPre, you can keep this\n")
                f.write("web: WEB             # temporary directory where piwigo thumbnails are written\n")
                f.write("\n")
                f.write("# remote-web:        # in HOME/.piwiPre.ini: the remote directory for thumbnails\n")
                f.write("# remote-album:      # in HOME/.piwiPre.ini: the remote directory for albums\n")
                f.write("# remote-user:       # in HOME/.piwiPre.ini: the ssh user\n")
                f.write("# remote-pwd:        # in HOME/.piwiPre.ini: the password of ssh user\n")
                f.write("# remote-host:       # in HOME/.piwiPre.ini: the ssh host\n")
                f.write("# remote-port:       # in HOME/.piwiPre.ini: the ssh port\n")
                f.write("# remote-incoming:   # in HOME/.piwiPre.ini: ssh incoming directory\n")
                f.write("\n")
                f.write("# piwigo-user:       # in HOME/.piwiPre.ini: CHANGE THIS to the name of the piwigo user\n")
                f.write("# sql-user:          # in HOME/.piwiPre.ini: CHANGE THIS to the name of the sql user\n")
                f.write("# sql-pwd:           # in HOME/.piwiPre.ini: CHANGE THIS to the pwd of sql user\n")
                f.write("# sql-host:          # in HOME/.piwiPre.ini: CHANGE THIS to the sql host\n")
                f.write("# sql-port:          # in HOME/.piwiPre.ini: CHANGE THIS to the sql port\n")
                f.write("# sql-database:      # in HOME/.piwiPre.ini: CHANGE THIS to the sql database\n")
                f.write("\n")
                f.write("enable-rename: true               # This value is usually OK\n")
                f.write("enable-rotation: true             # This value is usually OK\n")
                f.write("enable-metadata: true             # This value is usually OK\n")
                f.write("enable-conversion: true           # This value is usually OK\n")
                f.write("trust-date-in-filename: true      # This value is usually OK\n")
                f.write("\n")
                f.write("enable-thumbnails: true           # keep true\n")
                f.write("enable-thumbnails-delete: true    # keep true\n")
                f.write("enable-remote-web: true           # keep true: remote directory\n")
                f.write("enable-remote-album: true         # keep true: remote directory\n")
                f.write("enable-database: true             # keep true\n")
                f.write("enable-auto-configuration: true   # This value is usually OK\n")
                f.write("enable-pwg-representative: true   # keep true\n")
                f.write("\n")
                f.write("# End of file\n")
            LOGGER.msg(f"Generated '{ini_file}' for a remote usage")
            LOGGER.msg("")
            LOGGER.msg(f"your HOME is '{os.path.expanduser('~')}'")
            LOGGER.msg("")
            LOGGER.msg("The following value SHOULD be set in HOME/.piwiPre.ini:")
            LOGGER.msg("   piwigo-user:         # SET the name of the piwigo user")
            LOGGER.msg("   sql-user:            # SET the name of the sql user")
            LOGGER.msg("   sql-pwd:             # SET the pwd of sql user")
            LOGGER.msg("   sql-host:            # SET the sql host")
            LOGGER.msg("   sql-port:            # SET the sql port")
            LOGGER.msg("   sql-database:        # SET the sql database")
            LOGGER.msg("")
            LOGGER.msg("   remote-web:          # SET the remote directory for thumbnails")
            LOGGER.msg("   remote-album:        # SET the remote directory for albums")
            LOGGER.msg("   remote-user:         # SET the ssh user")
            LOGGER.msg("   remote-pwd:          # SET the password of ssh user")
            LOGGER.msg("   remote-host:         # SET the ssh host")
            LOGGER.msg("   remote-port:         # SET the ssh port")
            LOGGER.msg("   remote-incoming:     # SET ssh incoming directory")
            LOGGER.msg("")
        else:
            raise PwpConfigError(f"Illegal argument '{arg}' for reset-ini")

    def run(self):
        do_exit = False  # exit is delayed until we have managed all --flag that generate exit

        # We do it HERE because we want to have read the 1st .ini in cwd

        ACTOR.configure(self.parser_config)
        LOGGER.configure(self.parser_config)

        # done automatically by argparse
        # if self.config['help']:
        #     self.parser.print_help()
        #     do_exit = True

        if self.parser_config['version']:
            version = PwpVersion()
            # LOGGER.msg(f"current version: '{version.help}' ")
            print(f"current version: '{version.help}'")
            do_exit = True

        if self.parser_config['licence']:
            licence = PwpLicence()
            licence.print()
            do_exit = True

        if self.parser_config['reset-ini']:
            self.reset_ini(self.parser_config['reset-ini'])
            do_exit = True

        if self.parser_config['dump-config']:
            self.dump_config(self.parser_config['dump-config'], self.parser_config)
            do_exit = True

        # -------------------------------------------------------------
        # SQL CONNECT !
        #
        # has already been done in ACTOR.configure

        _conn, first_album, cause = ACTOR.connect_sql(self.parser_config)
        if self.parser_config['test-sql']:
            LOGGER.msg("Testing SQL ")
            LOGGER.msg(f"sql host  : '{ACTOR.sql_host}'")
            LOGGER.msg(f"sql port  : '{ACTOR.sql_port}'")
            LOGGER.msg(f"sql user  : '{ACTOR.sql_user}'")
            LOGGER.msg(f"1st album : '{first_album}'")
            if first_album:
                LOGGER.info(f"test-sql OK: 1st album = '{first_album}'")
            else:
                LOGGER.info(f"SQL ERROR: {cause}")  # pragma: no cover
            do_exit = True
        # -------------------------------------------------------------e
        # SSH CONNECT !
        #
        # has already been done in ACTOR.configure

        remote, uname, cause, is_error = ACTOR.connect_ssh(self.parser_config)

        if self.parser_config['test-ssh']:
            LOGGER.msg("Testing ssh ")
            LOGGER.msg(f"remote host : '{ACTOR.remote_host}'")
            LOGGER.msg(f"remote port : '{ACTOR.remote_port}'")
            LOGGER.msg(f"remote user : '{ACTOR.remote_user}'")
            LOGGER.msg(f"uname       : '{uname}'")
            if remote:
                result = ACTOR.remote_ls(".")
                LOGGER.info(f"test-ssh OK: ls -l       : '{result}'")
            else:
                LOGGER.info(f"Cannot ssh : {cause}")  # pragma: no cover
            do_exit = True

        if self.parser_config['test-sftp']:
            if not remote:  # pragma: no cover
                LOGGER.msg("sftp test OK : skipped because no remote configuration")
                LOGGER.info("sftp test OK : skipped because no remote configuration")
            else:
                dummy = "dummy.txt"
                dummy_timestamp = ACTOR.build_timestamp(dummy)
                base = os.path.basename(dummy)
                LOGGER.debug("Testing sftp")

                dst = self.parser_config['remote-web']
                ACTOR.remote_put(dummy, dst)
                result = ACTOR.remote_ls(dst)
                if base not in result:  # pragma: no cover
                    LOGGER.info(f"sftp failed      : '{result}'")
                    LOGGER.msg("sftp test failed")
                else:
                    remote_stamp = ACTOR.timestamp_from_ls(result[base])
                    if remote_stamp == dummy_timestamp:
                        LOGGER.info(f"sftp test OK          : '{result[base]}'")
                        LOGGER.msg("sftp test OK")
                    else:
                        LOGGER.info(f"sftp set time failed      : '{dummy_timestamp}'  '{remote_stamp}")
                        LOGGER.msg("sftp test failed")

                ACTOR.remote_delete(dst + '/' + dummy)
            do_exit = True

        if do_exit:
            LOGGER.msg("Exiting due to cmdline options")
        else:
            with tempfile.TemporaryDirectory() as tmp_dir:
                self.parser_config['tmp_dir'] = tmp_dir

                if self.parser_config['verify-album']:
                    if self.parser_config['triage']:  # pragma: no cover
                        LOGGER.warning(f"removing target --triage {self.parser_config['triage']} " +
                                       "because --verify-album not empty")
                        self.parser_config['triage'] = None

                    to_verify = self.parser_config['verify-album']
                    if '*' in to_verify:
                        to_verify = [f for f in os.listdir(self.parser_config['album']) if os.path.isdir(f)]

                    for item in to_verify:
                        item_father = os.path.dirname(item)
                        if item_father == "":
                            father_path = self.parser_config['album']
                        else:
                            father_path = self.parser_config['album'] + '/' + item_father
                        cur_dir = os.path.basename(item)
                        p_config = self.combined_config(father_path, self.parser_config, caller="verify-album")
                        new_base = self.get_base_from_dir(cur_dir, p_config)
                        recursive = self.parser_config['recursive-verify-album'] is True
                        dir_to_verify = PwpDirEntry.open(self.parser_config['album'] + '/' + item,
                                                         config=p_config,
                                                         context='album')
                        self.run_stage_dir('album', p_config, dir_to_verify, new_base, recursive=recursive)

                else:
                    target_dir = PwpDirEntry.open(self.parser_config['triage'],
                                                  config=self.parser_config,
                                                  context='local')
                    self.run_stage_dir('triage', self.parser_config, target_dir, '', recursive=True)

            LOGGER.msg("End of processing")

        LOGGER.end()
        # print_open_file_descriptors()
        os.chdir(self.initial_cwd)


def pwp_init(arguments=None):
    """used for tests, when the test harness in test_pwp needs to use the ssh connection
    initializes PwpMain"""

    main = PwpMain(arguments)
    return main


def pwp_run(main: PwpMain):
    """used for tests, when the test harness in test_pwp needs to use the ssh connection"""
    if main is None:
        raise PwpError("main is None")
    main.run()
    return main.parser_config if main is not None else None


def pwp_main(arguments=None):

    main = PwpMain(arguments)
    if main.parser_config is None:
        return None
    return pwp_run(main)


# def print_open_file_descriptors():
#     fd_types = {
#         'REG': stat.S_ISREG,
#         'FIFO': stat.S_ISFIFO,
#         'DIR': stat.S_ISDIR,
#         'CHR': stat.S_ISCHR,
#         'BLK': stat.S_ISBLK,
#         'LNK': stat.S_ISLNK,
#         'SOCK': stat.S_ISSOCK,
#     }
#     print("Open file descriptors between 0 and 100\n")
#     for fd in range(100):
#         try:
#             s = os.fstat(fd)  # noqa
#         except OSError:
#             continue
#         msg = f"File descriptor {fd:2} struct= {s}  "
#         for mode, mask in fd_types.items():
#             if mask(s.st_mode):
#                 msg += " " + mode
#         print(msg, flush=True)
#     print("end\n")
#     return


if __name__ == "__main__":
    pwp_main(sys.argv[1:])



# file generated by pwpImageEmbedder from pwpLogo.png

import io
from piwiPre.pwpImageEmbedder import PwpBioImage

pwpLogo_png = PwpBioImage(io.BytesIO(b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x002\x00\x00\x00\x1c\x08\x06\x00'	# noqa
                                     b'\x00\x00#\x99\xec\xf7\x00\x00\n)IDATx\x9c\xc5\x98{\x8c\\\xd5y\xc0\x7f\xdf9g\xee'	# noqa
                                     b'<w\xd6\xbb\xeb56\xebW\x9d`(\xc6`\x92\xb8\xb4\x81(\xde\x84\x06\x1a\xd2\x02Fr\xd2'	# noqa
                                     b'$U\xd2\xa8U\xab>B\xa3\x8a\xb6\x12\xaal\xa7\x02"\x84\x14\xa9y\x82\xd2\xd0T\x91'	# noqa
                                     b'\xdd\xae\x83B\xd2\x00\r\x01\x0c1uE\xc3\xcb\xa1k\x1c\x83\xe3]\xbc~\xaew\xd7\x9e'	# noqa
                                     b'\x9d\xc7\xbdw\xee9_\xff\x98]?\xc0&`\\\xf5\x93\x8e\xe6j4\xf7;\xdf\xef{\x9d\xef'	# noqa
                                     b'\x8c\xf0\xff$\xfa\xb7|\x99\x88\xbfz%\xbb(\xa3\xd8\xe3\x04\x0f\x08\xa0\x04,\xaeq'	# noqa
                                     b'([\x96\x7f\xcd\xf9\xd8\xdc\xe6\xee\x0e\xf7\xfc*}\xf2\x8e\r\x02\x11P\xfd;\x96'	# noqa
                                     b'\x93c\x01\x19\x810\xa3\xd7\x15:kV\xb2\xb8\xb3\xc0`\xcdw\x92zX\xfc\xd2\xdaG\xc2'	# noqa
                                     b'\xca\x0f^o\xf4\x04F\xe7\xf3\x85\xef}5\\\xb5\xfd/\x8d\x14\xdd\x0e|\xf6y\xc0\xa0'	# noqa
                                     b"\xa4\xb4xN\xbeB\xf2z;\xdcy\x81XOD\xcc\xa3X\x96\x10\x00'\x1ds\x92\x18\x1a\xf1I"	# noqa
                                     b'\xeb\xf2@\xde\x82\x06\xf0J\x10\x08{\xb7\x99CE\x8f\x04\x0f\xc6\x80\x06T\x1c\xf6'	# noqa
                                     b'\xd0\x8b&(X\xd1+\xc8\xf1\x14\x01\xe8\xf8\xe4\xe3\xc0\x90\xae\xc7\xc9F\xb2\xf3'	# noqa
                                     b'\x02r\x8a\x18\xc0\x92\x10\x08\x12\xf0j|\xa2\xd8\xe5\xab`\xf9\x07\xa1\xd0\x0b'	# noqa
                                     b"\x8d\xc3\xb0\xf3'\xf8\x91W\xb0%@\x90b\x11Y\xb5\xe3N\xda\xcf\x9d\x9e\x1a\n\x0cD`"	# noqa
                                     b'\x0b@\xdb\xeb\xccW\x19\x01\x87P:\x93\x01\xe7\x0b\x04\x84\x80\x11C\xaahq\x8e\xb1'	# noqa
                                     b'\x9f\xfb\x16\xfbJ\xef\xe5\x85\xdd\xafQ\x9fhP.\x15\xb9j\xdd_3\xff\xf0\xa3\x84'	# noqa
                                     b"\x7f\xfb\x0b\x8c\t\x00D\x05\x88\xe4\x0c\x19\xae\xdaYB'\xbc\x8aA1t\x8a\xe9\xff"	# noqa
                                     b"\x10DE\x08B\xc8\x950\xb7n\xe3\x81g'xx\xcb\x17\xe9\xe9\xaeR\xa9T\xe8\xe9\xe9\xe5"	# noqa
                                     b"?\x1f\x7f\x98\x0f\xfd\xde'\xb8\xfe\xcf\xfe\x9d\xf0\x8f\x1f\xc5\x14:\xa9\x84\xea"	# noqa
                                     b'\xdb\xd8\xe7\xccum\xde\xf4\x1dUQ\x1d\xb2\xbau\xbdS\x1d\xb2\xaa\xeb\xcf\xfe{c'	# noqa
                                     b"\xa1\x110k\xef\xe0\xb1=\xf0\xf5\xbbn'\x8e\x13F\xf7\xedg\xc7K\xc3\xfc\xfc\x7f"	# noqa
                                     b'\x86Y\xbd\xfa}<\xf6\xc0w\xf8Y\xbc\x02s\xed\x9f\x12\xea\x01\xcc\xf9\xf1\xe5\x1b'	# noqa
                                     b'\xb4\xa8\xae7<\x89\xd90H\x10\x91\x00o\x08\xa5\xe8\xd6\xf5\x965\x04d\xe3\x8c+'	# noqa
                                     b'\x97@6\n\xdd\x05\xb2\x8bn\xe6{\xb7\xdfCT,S\xaf\xd7QU\x8c1\x8c\x8d\x8d\xb1i\xd3f'	# noqa
                                     b'\xd6\xdet#?y\xe8A\xde\xf7\xc9\xcf\xc2c\xdf\xecD\xe4|\x80\xa8"l\x192\xf4\x0f\x8b'	# noqa
                                     b'\x0cn\xccD6\x06 t\xa0\xfe\xeb\x02\x8e\x8c\xae\xf2\xd3a\xbe\xadF\xe3\xf4\xf7\x0c'	# noqa
                                     b'\x8b\xfb\xf0\xa8\x0cn\xcc\x00t\xfdz\xa7\x1b6\x04\xfey\ra\xc7(f`\x11\x13M\xc3'	# noqa
                                     b'\xf8\xe1\x83DQ\x1e\x99\xc9}U\xa5R\xa9P\xab\xd5\xd8;2\x8a\xc9\x97h\xf8\n\x95j'	# noqa
                                     b'\x11m\xb5\x103\xdbx\xdf\x01\x88\x08\n\xeb<\x80DBH~\xb4\xb2\xbdk\xe2\xba\xf8`r'	# noqa
                                     b'\xdd\x91-\xcf\xae\xce\xb5\xa4\xdbz\xf0\x06\xda\xd1\xc1\xd6\xb1G\xee\xfby\xe5'	# noqa
                                     b'\xc2\xc2#v\xc5\xdc\xef\x8a|t\x0f\x1b7\xa2\xf7~\xcc`Eh\xb7\x88\x9c`\x9d\xa3\x9d'	# noqa
                                     b'\xc4\xa7\xe5~\x08\x01\xe7\x1c\xd3\xd35\xca6O\xce\x02Y\xc6\x99\xea\xfc\x9c@T\xd5'	# noqa
                                     b'\xd0\xfe\xe1o\xa4\xbbj7\xb4\x0e6\x7f\xe7\xf0w_}O1\xb1\x12\x1fOi6\x132\r\xc1\x18'	# noqa
                                     b'\t\x1aT\x8cH1\x8c%W\xc9Hz\x95\xdf9}[\xfa\xdf\xdf~0\xb7z\xdeWD>\xf6\x8c\xff\x02'	# noqa
                                     b'\x9e\xa3\x07\xe91S\xfc\xfaeW\xb2\xed\xf1\x87\xe8\xed\xeb\xa7\xddn\x03\x9d\xa8'	# noqa
                                     b'\x94+\x15\xc6\x0f\x1f\xe4\xe2U\xbfE\xbe\xb9\x87Pkc\xba,\x8436\xa2\xb7\x07r\xf4'	# noqa
                                     b'\xfb_{\xd64\xb9\xd2\xd6\x95f-\xa6\x19\xa74\x8cdb\x11\xcd\x89\xb1\x18\x83b\xc4\t'	# noqa
                                     b'\x806}[\x1bG\xd2 \x87(\xeb\xe1\xf6\xa7\xf4\x95\xc6\xa7\xea\xdb\xff\xe9A\xf3\xcc'	# noqa
                                     b'?\x14\xd9=\x02\x8f\xdf!\x9f\xfb\xf3\xfby\xf9\xc5g\x98\x988J\xb9\xab\x0b\x11\xc1'	# noqa
                                     b"g\x9e\xfa\xf1)\xe6]z9\xb7\xac\xfd]\xf4\xfe\x0f \x8ew\x9aQ'Aj;k\x03\x16\xa3\x9e"	# noqa
                                     b'\x90a\xc5\x9a\xc8\x08\xdai\x02r\xea&\x9dg\x11\x10\xb1bp\xe8T\xad\x19\x98\x0c'	# noqa
                                     b'\xb6:\xeeoJ\x16~\x819s\xff\x03\xf3\xf4\xbf\xca\x92\x0b.\xe7\x9eo|\x9bo\xddw//'	# noqa
                                     b'\xefx\x9e$\x89)U\xab|\xe0\xb7o\xe0\x8f>\xfbiz\x1e\xfd<\xfa\x8b\x9f!\x15s^\xa2'	# noqa
                                     b'\x01\xe0\xc4\xca\x13\xc5B\xf4\x89F\x9c\x18:\xf3\xcc[\x13E\x8c\x11K\xde\xeat\xb3'	# noqa
                                     b'\x11\xc2\xb0\x98\xc6\xc2\x1be\xde\x87\x96\x90\xff\xf1\xed,\xde\xbb\x8d/~\xf2o'	# noqa
                                     b'\x98\xf8\xcc-\xd4SeNA\xb4;\x1dQ6\xdf@\x18\xde\x86\xa9X\xf0o\x0b"\xa0\x04\xc2'	# noqa
                                     b"\x99-t\xaa\xfa\xaah'\x87\xe5\\*O\x11\x11c]\x01\xd2\xfd\xe3\x1c\xac_\xca\x9c\xd5"	# noqa
                                     b'wP\x9a|\x8ah\xf3-\xf4\xb9\x88>\x89\xa0y\\\xa8\x1f\x17,\x98>\x07Y\x06\xf6\x14=Bg'	# noqa
                                     b'\xd6\xeaLo\x10\xc2\xc9\x19\xcd\x18h\x87\x08\x0b\xde\xb9\xd2\xd6\xad\xb7;FF\x9c'	# noqa
                                     b'\xea\xd20\xd3eqb\xccKY\x08\x9c\x1b\xc5)<\n&\xef\xd0F\x8d\x89\x9d9\x8e\xcd\xb9'	# noqa
                                     b'\x89\xd2\xd2\xeb)0J)\x1e\xc6\xf9\x17j$\xbcJL\x12\x92\xacd\xf2\xa6x\xa2\xad\x89'	# noqa
                                     b'\x98\x10B \t\xd3FI\x01K\x9e*FLP\xc5$a\x8a2E\n\xf9\xc8\xeeK\x9e\x1e\xec\xb4\xff'	# noqa
                                     b'\x8c?\x04\xd5!\x0b\xeb\x82\xfc\xf2\xae\xbb\xaeP\xd5\xe7\x8c1V\xdf\xce\xa8\xf0&"'	# noqa
                                     b'\xa2h\x16\x08\xc1\xa0\xae@\xae\xbb\xa0\xa5~\xea\x85\xfeh{iq\xf1Gv\xd9\xdc\x87D'	# noqa
                                     b'\xae\xdb{\xfaK\xa0\xc7\x87\xae\x89\xc7\xb9\xbc\x90o\xefc\xe0\xc2GD\x06;\xe7\xd5'	# noqa
                                     b'\xd8\xa6\xcf4\x7fQ\xff\xe3VM\xe7\x15\xdfU~\xb0\xd4g_lO\xb38w\xa1\xfd\xa1T\xd7'	# noqa
                                     b'\xed\x04\x90\xc9/}\xa9{*\x84]\x91s\xf3\xd3,Sy\x87\x91y\xbdq\xa2J\xf0\x01QC\xb9T'	# noqa
                                     b'\xa0\xd0\x1d\x11\xe6\x98z\xe5\xdd\x95\x7fqW,\xbb\x83\xe6\x01\xf1\x07\xfc\xcd'	# noqa
                                     b'\xcd\x03\xcd\xb5\xed}\xcdA\x9dVz\xe6\x96i_ln\xcd\xaf\\\xb9\xc9\xbf8\xfc\xcdd'	# noqa
                                     b'\xb8y\xcb\xd4\xc84\xf9\xf9yz\xdeU\xa5\xb1\xaf\x89\xb4@\x17\xd9\xe9\xea\xd5\xfd'	# noqa
                                     b'\xb7R}\xcff\x01\xd8s\xe7\x9d\xcf\x97\xf3\xf9+\x1bi\x1a\xe4W\xcc_g\x15\xe5\xf47_'	# noqa
                                     b'\x1f\\\x01U\r\x9a\xa9\x1a\x15\xdb\xd7\xdfE{\xbeL\x15\x07J\x86\xddiw\xed\xb5i'	# noqa
                                     b"\xa6\xe3\x18W\xb4\xed\xf2\xdc\x92\xad\xac\xea\xda\xdf\x18m\x8a\x1b\xf3\x0b'\x8e"	# noqa
                                     b'\xd5}\xf7\x8anQU\x99\xde]\x0f\xa8\x12\xda\xaa=\x97v\xbb\xee\\\x91\xb8/<\xde\x99'	# noqa
                                     b'\xb5D\xc6\xad1\xb3\x15\x7fN\x1c\xe2\x84\xd0\x0e\xa8\x9fi\x1a\x86\xce\xe81\xab'	# noqa
                                     b"\xae3\x91w\xce#A\xc7'j\xc1\x1c\n=\xf1H\x9dha1caN\xba\xa3\xbc59\x93Kk)S?\x9d\\"	# noqa
                                     b'\x84B:\xc7\xf8\xdee\xbd6>\x12\xd3\xda\xdf\xc2D\xc6\xaaB\xbe\x9a\x83\x82\t{\xb7'	# noqa
                                     b'\x1f\xc8\x06\xae\xbe\xe0\xc3n\xc6Y\xe3F\x04\x119\xb7"Q(\x0e\x14\xb1\x05\x8boy'	# noqa
                                     b'\xb2FFV\xcf\xf0\xb1?3\x98"\xc6\x1a\xdb\xb5\xfa\xbd\x8a\xc9\xd1\xdc\xf5\xb2\xab.'	# noqa
                                     b'7\xc4\x87R\x1a#\r\xa2\x9e\x88\xdc@^Md4\xabe\xb6\xb6\xab\x86\x8f=&\xdf\tyH5\x14'	# noqa
                                     b'\x07\x8a!\xd9\xdbt\x8bW\xcc\x8b\xc2\x02\xb7\xc9\x01\xa8\xea\x01AB\xf0\x9a\xc9'	# noqa
                                     b"\x0c\xd9\x8c'e\xa6\x9b\x89\x08r\xd63F\xa01\xda\xc0F\x16W\xb6\xd8\x8a#\xea\x8d"	# noqa
                                     b'\x10\x01\x9f\x04\xb2\xfa)`\x01\xb4\xdd\xa6\xb8d\x11\xe5\x8b\x96\x0b\x1a(,\x98'	# noqa
                                     b'\xcf\xe4\x13OP\xbd(\x87\xeb\xca\x91M\xb7\x89\x0f\xc5\x92\xd53\xd1\xa0\x18g\xb0y'	# noqa
                                     b'K\x08\x1a\xb2f&\x0b.\x9fkJ\xdd\x05\xd3\x10w wY\xf9kv\xf9\xa7\xefv\x00\xd6\xb9'	# noqa
                                     b'\x1f\xd8\x82\xb9\xad\xbfR-\xa0@\x80\x90\x05|\xdb\xd3N=I\x9a\x91\x05\x1f\xc4J'	# noqa
                                     b'\x10#\x82`\xe0t0\x11!\xa4\x81\xa4\xe5\xd1\xf1\x041\x82\xc9\x1b\\\xc5\xe1*\x8e'	# noqa
                                     b'\xa8/B\x10B[I\xa7\x1a\x14\x17/\x00\x0cY\xbd\x81\xab\x96\xc8/\\L2\xfe\n\xc9\x94'	# noqa
                                     b'\x92N&\x98\xc8 N0\x18\x02\xaa>\xc9B9_\xb0]\xab{)\xae\xec\xdaN#l\xaa^\xbbz\xb3'	# noqa
                                     b'\xc8e\x93\xf0\x07\xc8\xcc9\xa8\x93O\xdfwc\xa9\xb7\xb4\x8c\xb6\xd7,\xf5\x15m\xeb'	# noqa
                                     b'\x02\xdf\xca.\xf4IX\xe4\xeb~\xa9\x8d\xb5\xcf\xb6 \x9eNi\xb6\x12\x82\xaa\x17gT'	# noqa
                                     b'\xcc\t\xa8\xd9\x18\x9e\x94\x00\xea\xb5Sz\xb3`\xe5\x1c\xae$D\xfd}D\x0b\xde\x8d'	# noqa
                                     b'\xed\xeaAr%&\x9fz\x9a\xa88Nc\xacMH\xfc\xac\x1e\r\x99z\x87q}\x03]\xd8\xa5\xf9'	# noqa
                                     b'\xe7\x8b\xd7\xf4\xfd\xbd\xe4n~\xf8DV\x0f\rY\xf3\xf1u~v\xdb\xb3_\x08,h\xf6L\x1f'	# noqa
                                     b'\xc7F\xaf\x88G\x9ak\xe2\xa3\xe9`{"\xb92\x8aM9\xd43\x9a\xf5\x84$m\xa3\x827\xd6('	# noqa
                                     b'\x1d\xb0\x93\xa3\xce\xeb\xc1\x82v\x8eA\xf5\xd8\xbc\x90\x9f\xdf\xab\x85\x85\xf3'	# noqa
                                     b'\t\xf18\xe9xS\x9bc-\xc4\xa2\xeaQ\xab\xe2z\xe7u\xc1\x02w\xa8\xba\xb2z7K\x7f\xff'	# noqa
                                     b'\xab"\xd2VT\xd8\xba\xc1\xb2f\x83\x9f\xad\xeb\x13\xdb\xe8\xd0\x90\xa5\xbf\xff'	# noqa
                                     b'\x94m\x9f\x84\xaf\xafP\xd9\xb2.\x9c\x06)\xa0\xe1\xa9E~\xdf\xfekZ\xfb[\xd7\xa6'	# noqa
                                     b'\x13\xe9\xfb\xd3\xa9\xf4\xe2r\xe6$\xabg\xb4\x1a\t\xad\xb4\xad\xe2D\xc5\x88\x01P'	# noqa
                                     b'\xa5\xf3OH\xe8<\xe9L\x04E\x112\x8f1\x90+\xe7\xd1Tq9\x8b\xcbY\n\xd5\x08\xdfk\xa6'	# noqa
                                     b'\xba/\xe9\xba\x97K~\xed\xcb"W\x1f\xe9\xe8\x1a\xb2"\xeb\xde0\xa4\xbd\xa5^\xab'	# noqa
                                     b'\xaa\xc2\x96-\x86\xfeaap\xa3\x97S\xc0T\xd5\xc0C\x17\xa7{\x8e\xfff6\x9e^\x13\x1f'	# noqa
                                     b'm\xbd\xdfO\xb4/\tS\x9ef=\x06\xc09K.r\xe4\xf2\x16\x89\x0cj\xc1\x8b\xe2\x8d\x06'	# noqa
                                     b'\xe3$\x11kZ\x88\xb6Ld[\x923\xd3\xb9.7Z]P~\x82\xe5s\x1f\x10\xf9\xc8\x81Y\x00X'	# noqa
                                     b"\x17\xe4,-\xe7-\xdd\xfcg\xc2w\xc2\x0b\xb3\xf7\xfa'\x9f\x04\x11\xc9\x80\x97g\xd6"	# noqa
                                     b'\xfd\xaa\x9a\xa7\xf6\xc0G\xe2\xe1\xda\x9f\xb8C\xadA\x9f\x05\xeb\xca\xee\x80-'	# noqa
                                     b'\xd9_\xba\xa2\xdb\xed\xaa\xb9W]\xd9\x8e\xd9\x92;\xca\xdc\xc21p\xc7!\xaa\xc3@'	# noqa
                                     b"\x13.IL\xde\xa4\x9a\x9e\xb4u\xeb\xd6\xf5nM'\x85\xdetT\xfe_P,\xf7:\x15\x95_I\x00"	# noqa
                                     b'\x00\x00\x00IEND\xaeB`\x82'),	# noqa
               img_format="png",	# noqa
               filename="pwpLogo.png")	# noqa
